"""
Authentication class definitions.
"""
from dataclasses import dataclass
from typing import Optional, Union, List

from weaviate.warnings import _Warnings

SCOPES = Union[str, List[str]]


@dataclass
class AuthClientCredentials:
    """Authenticate for the Client Credential flow using client secrets.

    Acquire the client secret from your identify provider and set the appropriate scope. The client includes hardcoded
    scopes for Azure, otherwise it needs to be supplied.
    Scopes can be given as:
      - List of strings: ["scope1", "scope2"]
      - space separated string: "scope1 scope2"
    """

    client_secret: str
    scope: Optional[SCOPES] = None

    def __post_init__(self) -> None:
        if self.scope is None:
            self.scope_list: List[str] = []
        elif isinstance(self.scope, str):
            self.scope_list = self.scope.split(" ")
        elif isinstance(self.scope, list):
            self.scope_list = self.scope


@dataclass
class AuthClientPassword:
    """Using username and password for authentication with Resource Owner Password flow.

    For some providers the scope needs to contain "offline_access" (and "openid" which is automatically added) to return
    a refresh token. Without a refresh token the authentication will expire once the lifetime of the access token is up.
    Scopes can be given as:
      - List of strings: ["scope1", "scope2"]
      - space separated string: "scope1 scope2"

    """

    username: str
    password: str
    scope: Optional[SCOPES] = None

    def __post_init__(self) -> None:
        if self.scope is None:
            self.scope_list: List[str] = []
        elif isinstance(self.scope, str):
            self.scope_list = self.scope.split(" ")
        elif isinstance(self.scope, list):
            self.scope_list = self.scope


@dataclass
class AuthBearerToken:
    """Using a preexisting bearer/access token for authentication.

    The expiration time of access tokens is given in seconds.

    Only the access token is required. However, when no refresh token is given, the authentication will expire once
    the lifetime of the access token is up.
    """

    access_token: str
    expires_in: int = 60
    refresh_token: Optional[str] = None

    def __post_init__(self) -> None:
        if self.expires_in and self.expires_in < 0:
            _Warnings.auth_negative_expiration_time(self.expires_in)


@dataclass
class AuthApiKey:
    """Using the given API key to authenticate with weaviate."""

    api_key: str


OidcAuth = Union[AuthBearerToken, AuthClientPassword, AuthClientCredentials]
AuthCredentials = Union[OidcAuth, AuthApiKey]



from dataclasses import dataclass
from typing import (
    Dict,
    List,
    Literal,
    Optional,
    Protocol,
    Union,
    overload,
)
from typing_extensions import TypeVar

from pydantic import BaseModel, Field

from weaviate.collections.classes.config import DataType

N = TypeVar("N", int, float)


@dataclass
class AggregateInt:
    """The aggregation result for an integer property."""

    count: Optional[int]
    maximum: Optional[int]
    mean: Optional[float]
    median: Optional[int]
    mode: Optional[int]
    sum_: Optional[int]


@dataclass
class AggregateFloat:
    """The aggregation result for a float property."""

    count: Optional[int]
    maximum: Optional[float]
    mean: Optional[float]
    median: Optional[float]
    mode: Optional[float]
    sum_: Optional[float]


@dataclass
class TopOccurrence:
    """The top occurrence of a string property."""

    occurs: Optional[int]
    value: Optional[str]


@dataclass
class AggregateStr:
    """The aggregation result for a string property."""

    count: Optional[int]
    top_occurrences: Optional[List[TopOccurrence]]


@dataclass
class AggregateBool:
    """The aggregation result for a boolean property."""

    count: Optional[int]
    percentage_false: Optional[float]
    percentage_true: Optional[float]
    total_false: Optional[int]
    total_true: Optional[int]


@dataclass
class AggregateRef:
    """The aggregation result for a cross-reference property."""

    pointing_to: Optional[str]


@dataclass
class AggregateDate:
    """The aggregation result for a date property."""

    count: Optional[int]
    maximum: Optional[str]
    median: Optional[str]
    minimum: Optional[str]
    mode: Optional[str]


AggregateResult = Union[
    AggregateInt, AggregateFloat, AggregateStr, AggregateBool, AggregateDate, AggregateRef
]

AProperties = Dict[str, AggregateResult]


@dataclass
class _AggregateReturn:
    properties: AProperties
    total_count: Optional[int]


@dataclass
class _GroupedBy:
    prop: str
    value: str


@dataclass
class _AggregateGroupByReturn:
    grouped_by: _GroupedBy
    properties: AProperties
    total_count: Optional[int]


class _MetricsBase(BaseModel):
    property_name: str
    count: bool = Field(default=False)
    type_: bool = Field(default=False)


class _MetricsStr(_MetricsBase):
    top_occurrences_count: bool = Field(default=False)
    top_occurrences_value: bool = Field(default=False)

    def to_gql(self) -> str:
        body = " ".join(
            [
                "count" if self.count else "",
                "type" if self.type_ else "",
                "topOccurrences {"
                if self.top_occurrences_count or self.top_occurrences_value
                else "",
                "occurs" if self.top_occurrences_count else "",
                "value" if self.top_occurrences_value else "",
                "}" if self.top_occurrences_count or self.top_occurrences_value else "",
            ]
        )
        return f"{self.property_name} {{ {body} }}"


class _MetricsNumber(_MetricsBase):
    maximum: bool = Field(default=False)
    mean: bool = Field(default=False)
    median: bool = Field(default=False)
    minimum: bool = Field(default=False)
    mode: bool = Field(default=False)
    sum_: bool = Field(default=False)

    def to_gql(self) -> str:
        body = " ".join(
            [
                "count" if self.count else "",
                "type" if self.type_ else "",
                "maximum" if self.maximum else "",
                "mean" if self.mean else "",
                "median" if self.median else "",
                "minimum" if self.minimum else "",
                "mode" if self.mode else "",
                "sum" if self.sum_ else "",
            ]
        )
        return f"{self.property_name} {{ {body} }}"


class _MetricsInt(_MetricsNumber):
    pass


class _MetricsFloat(_MetricsNumber):
    pass


class _MetricsBool(_MetricsBase):
    percentage_false: bool = Field(default=False)
    percentage_true: bool = Field(default=False)
    total_false: bool = Field(default=False)
    total_true: bool = Field(default=False)

    def to_gql(self) -> str:
        body = " ".join(
            [
                "count" if self.count else "",
                "type" if self.type_ else "",
                "percentageFalse" if self.percentage_false else "",
                "percentageTrue" if self.percentage_true else "",
                "totalFalse" if self.total_false else "",
                "totalTrue" if self.total_true else "",
            ]
        )
        return f"{self.property_name} {{ {body} }}"


class _MetricsDate(_MetricsBase):
    maximum: bool = Field(default=False)
    median: bool = Field(default=False)
    minimum: bool = Field(default=False)
    mode: bool = Field(default=False)

    def to_gql(self) -> str:
        body = " ".join(
            [
                "count" if self.count else "",
                "type" if self.type_ else "",
                "maximum" if self.maximum else "",
                "median" if self.median else "",
                "minimum" if self.minimum else "",
                "mode" if self.mode else "",
            ]
        )
        return f"{self.property_name} {{ {body} }}"


class _MetricsRef(BaseModel):
    property_name: str
    pointing_to: bool = Field(default=False)
    type_: bool = Field(default=False)

    def to_gql(self) -> str:
        body = " ".join(
            [
                "pointingTo" if self.pointing_to else "",
                "type" if self.type_ else "",
            ]
        )
        return f"{self.property_name} {{ {body} }}"


_Metrics = Union[
    _MetricsStr,
    _MetricsInt,
    _MetricsFloat,
    _MetricsDate,
    _MetricsBool,
    _MetricsRef,
]

PropertiesMetrics = List[_Metrics]


class SupportsReturningStr(Protocol):
    """An interface for the `returning` method of the `Metrics` class for string properties."""

    def returning(
        self,
        count: bool = False,
        top_occurrences_count: bool = False,
        top_occurrences_value: bool = False,
    ) -> _MetricsStr:
        """Define the metrics to be returned for a string property when aggregating over a collection.

        Arguments:
            `count`
                Whether to include the number of objects that contain this property.
            `top_occurrences_count`
                Whether to include the number of the top occurrences of a property's value.
            `top_occurrences_value`
                Whether to include the value of the top occurrences of a property's value.

        Returns:
            A `_MetricsStr` object that includes the metrics to be returned.
        """
        ...


class SupportsReturningInt(Protocol):
    """An interface for the `returning` method of the `Metrics` class for integer properties."""

    def returning(
        self,
        count: bool = False,
        maximum: bool = False,
        mean: bool = False,
        median: bool = False,
        minimum: bool = False,
        mode: bool = False,
        sum_: bool = False,
    ) -> _MetricsInt:
        """Define the metrics to be returned for an integer property when aggregating over a collection.

        Arguments:
            `count`
                Whether to include the number of objects that contain this property.
            `maximum`
                Whether to include the maximum value of this property.
            `mean`
                Whether to include the mean value of this property.
            `median`
                Whether to include the median value of this property.
            `minimum`
                Whether to include the minimum value of this property.
            `mode`
                Whether to include the mode value of this property.
            `sum_`
                Whether to include the sum of this property.

        Returns:
            A `_MetricsInt` object that includes the metrics to be returned.
        """
        ...


class SupportsReturningFloat(Protocol):
    """An interface for the `returning` method of the `Metrics` class for float properties."""

    def returning(
        self,
        count: bool = False,
        maximum: bool = False,
        mean: bool = False,
        median: bool = False,
        minimum: bool = False,
        mode: bool = False,
        sum_: bool = False,
    ) -> _MetricsFloat:
        """Define the metrics to be returned for a float property when aggregating over a collection.

        Arguments:
            `count`
                Whether to include the number of objects that contain this property.
            `maximum`
                Whether to include the maximum value of this property.
            `mean`
                Whether to include the mean value of this property.
            `median`
                Whether to include the median value of this property.
            `minimum`
                Whether to include the minimum value of this property.
            `mode`
                Whether to include the mode value of this property.
            `sum_`
                Whether to include the sum of this property.

        Returns:
            A `_MetricsFloat` object that includes the metrics to be returned.
        """
        ...


class SupportsReturningBool(Protocol):
    """An interface for the `returning` method of the `Metrics` class for boolean properties."""

    def returning(
        self,
        count: bool = False,
        percentage_false: bool = False,
        percentage_true: bool = False,
        total_false: bool = False,
        total_true: bool = False,
    ) -> _MetricsBool:
        """Define the metrics to be returned for a boolean property when aggregating over a collection.

        Arguments:
            `count`
                Whether to include the number of objects that contain this property.
            `percentage_false`
                Whether to include the percentage of objects that have a false value for this property.
            `percentage_true`
                Whether to include the percentage of objects that have a true value for this property.
            `total_false`
                Whether to include the total number of objects that have a false value for this property.
            `total_true`
                Whether to include the total number of objects that have a true value for this property.

        Returns:
            A `_MetricsBool` object that includes the metrics to be returned.
        """
        ...


class SupportsReturningDate(Protocol):
    """An interface for the `returning` method of the `Metrics` class for date properties."""

    def returning(
        self,
        count: bool = False,
        maximum: bool = False,
        median: bool = False,
        minimum: bool = False,
        mode: bool = False,
    ) -> _MetricsDate:
        """Define the metrics to be returned for a date property when aggregating over a collection.

        Arguments:
            `count`
                Whether to include the number of objects that contain this property.
            `maximum`
                Whether to include the maximum value of this property.
            `median`
                Whether to include the median value of this property.
            `minimum`
                Whether to include the minimum value of this property.
            `mode`
                Whether to include the mode value of this property.

        Returns:
            A `_MetricsDate` object that includes the metrics to be returned.
        """
        ...


class SupportsReturningRef(Protocol):
    """An interface for the `returning` method of the `Metrics` class for cross-reference properties."""

    def returning(
        self,
        pointing_to: bool = False,
    ) -> _MetricsRef:
        """Define the metrics to be returned for a cross-reference property when aggregating over a collection.

        Arguments:
            `pointing_to`
                Whether to include the collection names that this property references.

        Returns:
            A `_MetricsRef` object that includes the metrics to be returned.
        """
        ...


# type ignore all the overloads of __new__ here because mypy is not capable of understanding that the return type is a protocol (interface)
# that the class (type) implements since Aggregate inherits from all the returned protocols
class Metrics(
    SupportsReturningStr,
    SupportsReturningInt,
    SupportsReturningFloat,
    SupportsReturningBool,
    SupportsReturningDate,
    SupportsReturningRef,
):
    """Define the metrics to be returned based on a property when aggregating over a collection.

    Use the `__init__` method to define the name to the property to be aggregated on and
    its data type using `weaviate.classes.DataType` or `"cref"` if the property is a cross-reference
    then use the methods of this class to define the specific available metrics to be returned.

    See [the docs](https://weaviate.io/developers/weaviate/search/aggregate) for more details!
    """

    @overload
    def __new__(cls, property_: str, type_: Literal[DataType.TEXT]) -> SupportsReturningStr:  # type: ignore
        ...

    @overload
    def __new__(cls, property_: str, type_: Literal[DataType.TEXT_ARRAY]) -> SupportsReturningStr:  # type: ignore
        ...

    @overload
    def __new__(cls, property_: str, type_: Literal[DataType.INT]) -> SupportsReturningInt:  # type: ignore
        ...

    @overload
    def __new__(cls, property_: str, type_: Literal[DataType.INT_ARRAY]) -> SupportsReturningInt:  # type: ignore
        ...

    @overload
    def __new__(cls, property_: str, type_: Literal[DataType.NUMBER]) -> SupportsReturningFloat:  # type: ignore
        ...

    @overload
    def __new__(cls, property_: str, type_: Literal[DataType.NUMBER_ARRAY]) -> SupportsReturningFloat:  # type: ignore
        ...

    @overload
    def __new__(cls, property_: str, type_: Literal[DataType.BOOL]) -> SupportsReturningBool:  # type: ignore
        ...

    @overload
    def __new__(cls, property_: str, type_: Literal[DataType.BOOL_ARRAY]) -> SupportsReturningBool:  # type: ignore
        ...

    @overload
    def __new__(cls, property_: str, type_: Literal[DataType.DATE]) -> SupportsReturningDate:  # type: ignore
        ...

    @overload
    def __new__(cls, property_: str, type_: Literal[DataType.DATE_ARRAY]) -> SupportsReturningDate:  # type: ignore
        ...

    @overload
    def __new__(cls, property_: str, type_: Literal["cref"]) -> SupportsReturningRef:  # type: ignore
        ...

    def __new__(cls, property_: str, type_: Union[DataType, Literal["cref"]]) -> "Metrics":
        """Create a new `Metrics` object."""
        return super().__new__(cls)

    def __init__(self, property_: str, type_: Union[DataType, Literal["cref"]]) -> None:
        self.__property = property_
        self.__type = type_

    def returning(self, **kwargs) -> _Metrics:  # type: ignore
        """Define the metrics to be returned for a property when aggregating over a collection.

        If you're seeing this docstring then you've likely forgotten to define the data type of the property.

        Example:
            ```python
            import weaviate.classes as wvc
            wvc.Metrics("myProperty", DataType.TEXT).returning(count=True)
            ```
        """
        if self.__type == DataType.TEXT or self.__type == DataType.TEXT_ARRAY:
            return self.__str(**kwargs)
        elif self.__type == DataType.INT or self.__type == DataType.INT_ARRAY:
            return self.__int(**kwargs)
        elif self.__type == DataType.NUMBER or self.__type == DataType.NUMBER_ARRAY:
            return self.__float(**kwargs)
        elif self.__type == DataType.BOOL or self.__type == DataType.BOOL_ARRAY:
            return self.__bool(**kwargs)
        elif self.__type == DataType.DATE or self.__type == DataType.DATE_ARRAY:
            return self.__date(**kwargs)
        elif self.__type == "ref":
            return self.__ref(**kwargs)
        else:
            raise ValueError(f"Unknown type {self.__type}")

    def __str(
        self,
        count: bool = False,
        top_occurrences_count: bool = False,
        top_occurrences_value: bool = False,
    ) -> _MetricsStr:
        return _MetricsStr(
            property_name=self.__property,
            count=count,
            top_occurrences_count=top_occurrences_count,
            top_occurrences_value=top_occurrences_value,
        )

    def __int(
        self,
        count: bool = False,
        maximum: bool = False,
        mean: bool = False,
        median: bool = False,
        mode: bool = False,
        sum_: bool = False,
    ) -> _MetricsInt:
        return _MetricsInt(
            property_name=self.__property,
            count=count,
            maximum=maximum,
            mean=mean,
            median=median,
            mode=mode,
            sum_=sum_,
        )

    def __float(
        self,
        count: bool = False,
        maximum: bool = False,
        mean: bool = False,
        median: bool = False,
        mode: bool = False,
        sum_: bool = False,
    ) -> _MetricsFloat:
        return _MetricsFloat(
            property_name=self.__property,
            count=count,
            maximum=maximum,
            mean=mean,
            median=median,
            mode=mode,
            sum_=sum_,
        )

    def __bool(
        self,
        count: bool = False,
        percentage_false: bool = False,
        percentage_true: bool = False,
        total_false: bool = False,
        total_true: bool = False,
    ) -> _MetricsBool:
        return _MetricsBool(
            property_name=self.__property,
            count=count,
            percentage_false=percentage_false,
            percentage_true=percentage_true,
            total_false=total_false,
            total_true=total_true,
        )

    def __date(
        self,
        count: bool = False,
        maximum: bool = False,
        median: bool = False,
        minimum: bool = False,
        mode: bool = False,
    ) -> _MetricsDate:
        return _MetricsDate(
            property_name=self.__property,
            count=count,
            maximum=maximum,
            median=median,
            minimum=minimum,
            mode=mode,
        )

    def __ref(
        self,
        pointing_to: bool = False,
    ) -> _MetricsRef:
        return _MetricsRef(
            property_name=self.__property,
            pointing_to=pointing_to,
        )



from weaviate.proto.v1 import batch_pb2 as _batch_pb2
from weaviate.proto.v1 import search_get_pb2 as _search_get_pb2
from google.protobuf import descriptor as _descriptor
from typing import ClassVar as _ClassVar

DESCRIPTOR: _descriptor.FileDescriptor



import pytest

import weaviate

schema = {
    "classes": [
        {
            "class": "Label",
            "description": "a label describing a message",
            "properties": [
                {
                    "name": "name",
                    "description": "The name of this label",
                    "dataType": ["string"],
                },
                {
                    "name": "description",
                    "description": "The description of this label",
                    "dataType": ["text"],
                },
            ],
        },
        {
            "class": "Message",
            "description": "a message from written by a person",
            "properties": [
                {
                    "name": "content",
                    "description": "The content of the message",
                    "dataType": ["text"],
                },
                {
                    "name": "labeled",
                    "description": "The label assigned to this message",
                    "dataType": ["Label"],
                },
            ],
        },
    ]
}


@pytest.fixture(scope="module")
def client():
    client = weaviate.Client("http://localhost:8080")
    client.schema.create(schema)
    yield client
    client.schema.delete_all()


def test_contextual(client: weaviate.Client):
    # Create labels
    client.data_object.create(
        {"name": "positive", "description": "A positive, good, happy or supporting message."},
        "Label",
    )
    client.data_object.create(
        {"name": "negative", "description": "A negative, bad, sad or disrupting message."}, "Label"
    )

    client.data_object.create(
        {
            "content": "ALERT: So now we find out that the entire opponent “hit squad” illegally wiped their phones clean just prior to the investigation of them, all using the same really dumb reason for this “accident”, just like other people smashing her phones with a hammer, & DELETING THEIR EMAILS!"
        },
        "Message",
    )
    client.data_object.create(
        {
            "content": "I'm so happy, proud and excited to be a part of this community for the rest of my days."
        },
        "Message",
    )
    client.data_object.create(
        {"content": "thank you for reminding the world of our cause"}, "Message"
    )

    client.classification.schedule().with_type("text2vec-contextionary-contextual").with_class_name(
        "Message"
    ).with_based_on_properties(["content"]).with_classify_properties(
        ["labeled"]
    ).with_wait_for_completion().do()

    result = (
        client.query.get("Message", ["content", "labeled {... on Label {name description}}"])
        .with_additional(["id", "classification{basedOn, id}"])
        .do()
    )
    labeled_messages = result["data"]["Get"]["Message"]
    for message in labeled_messages:
        assert message["labeled"] is not None
        assert message["_additional"]["id"] is not None



from typing import Any, Dict, List, Union, cast

from weaviate.collections.classes.config import (
    _CollectionConfig,
    _CollectionConfigSimple,
    _PQConfig,
    DataType,
    _InvertedIndexConfig,
    _BM25Config,
    _StopwordsConfig,
    _MultiTenancyConfig,
    _Property,
    _ReferenceDataType,
    _ReferenceDataTypeMultiTarget,
    _ReplicationConfig,
    _ShardingConfig,
    _VectorIndexConfig,
    StopwordsPreset,
    VectorDistance,
    PQEncoderType,
    PQEncoderDistribution,
    _VectorIndexType,
    Vectorizer,
    Tokenization,
    _PQEncoderConfig,
)


def _collection_config_simple_from_json(schema: Dict[str, Any]) -> _CollectionConfigSimple:
    def _is_primitive(d_type: str) -> bool:
        return d_type[0][0].lower() == d_type[0][0]

    def _property_data_type_from_weaviate_data_type(
        data_type: List[str],
    ) -> Union[DataType, _ReferenceDataType, _ReferenceDataTypeMultiTarget]:
        if len(data_type) == 1 and _is_primitive(data_type[0]):
            return DataType(data_type[0])

        if len(data_type) == 1:
            return _ReferenceDataType(target_collection=data_type[0])

        return _ReferenceDataTypeMultiTarget(target_collections=data_type)

    return _CollectionConfigSimple(
        name=schema["class"],
        description=schema.get("description"),
        properties=[
            _Property(
                data_type=_property_data_type_from_weaviate_data_type(prop["dataType"]),
                description=prop.get("description"),
                index_filterable=prop["indexFilterable"],
                index_searchable=prop["indexSearchable"],
                name=prop["name"],
                tokenization=Tokenization(prop["tokenization"])
                if prop.get("tokenization") is not None
                else None,
            )
            for prop in schema["properties"]
        ]
        if schema.get("properties") is not None
        else [],
        vectorizer=Vectorizer(schema["vectorizer"]),
    )


def _collection_config_from_json(schema: Dict[str, Any]) -> _CollectionConfig:
    def _is_primitive(d_type: str) -> bool:
        return d_type[0][0].lower() == d_type[0][0]

    def _property_data_type_from_weaviate_data_type(
        data_type: List[str],
    ) -> Union[DataType, _ReferenceDataType, _ReferenceDataTypeMultiTarget]:
        if len(data_type) == 1 and _is_primitive(data_type[0]):
            return DataType(data_type[0])

        if len(data_type) == 1:
            return _ReferenceDataType(target_collection=data_type[0])

        return _ReferenceDataTypeMultiTarget(target_collections=data_type)

    return _CollectionConfig(
        name=schema["class"],
        description=schema.get("description"),
        inverted_index_config=_InvertedIndexConfig(
            bm25=_BM25Config(
                b=schema["invertedIndexConfig"]["bm25"]["b"],
                k1=schema["invertedIndexConfig"]["bm25"]["k1"],
            ),
            cleanup_interval_seconds=schema["invertedIndexConfig"]["cleanupIntervalSeconds"],
            index_null_state=cast(dict, schema["invertedIndexConfig"]).get("indexNullState")
            is True,
            index_property_length=cast(dict, schema["invertedIndexConfig"]).get(
                "indexPropertyLength"
            )
            is True,
            index_timestamps=cast(dict, schema["invertedIndexConfig"]).get("indexTimestamps")
            is True,
            stopwords=_StopwordsConfig(
                preset=StopwordsPreset(schema["invertedIndexConfig"]["stopwords"]["preset"]),
                additions=schema["invertedIndexConfig"]["stopwords"]["additions"],
                removals=schema["invertedIndexConfig"]["stopwords"]["removals"],
            ),
        ),
        multi_tenancy_config=_MultiTenancyConfig(enabled=schema["multiTenancyConfig"]["enabled"]),
        properties=[
            _Property(
                data_type=_property_data_type_from_weaviate_data_type(prop["dataType"]),
                description=prop.get("description"),
                index_filterable=prop["indexFilterable"],
                index_searchable=prop["indexSearchable"],
                name=prop["name"],
                tokenization=Tokenization(prop["tokenization"])
                if prop.get("tokenization") is not None
                else None,
            )
            for prop in schema["properties"]
        ]
        if schema.get("properties") is not None
        else [],
        replication_config=_ReplicationConfig(factor=schema["replicationConfig"]["factor"]),
        sharding_config=_ShardingConfig(
            virtual_per_physical=schema["shardingConfig"]["virtualPerPhysical"],
            desired_count=schema["shardingConfig"]["desiredCount"],
            actual_count=schema["shardingConfig"]["actualCount"],
            desired_virtual_count=schema["shardingConfig"]["desiredVirtualCount"],
            actual_virtual_count=schema["shardingConfig"]["actualVirtualCount"],
            key=schema["shardingConfig"]["key"],
            strategy=schema["shardingConfig"]["strategy"],
            function=schema["shardingConfig"]["function"],
        ),
        vector_index_config=_VectorIndexConfig(
            cleanup_interval_seconds=schema["vectorIndexConfig"]["cleanupIntervalSeconds"],
            distance_metric=VectorDistance(schema["vectorIndexConfig"]["distance"]),
            dynamic_ef_min=schema["vectorIndexConfig"]["dynamicEfMin"],
            dynamic_ef_max=schema["vectorIndexConfig"]["dynamicEfMax"],
            dynamic_ef_factor=schema["vectorIndexConfig"]["dynamicEfFactor"],
            ef=schema["vectorIndexConfig"]["ef"],
            ef_construction=schema["vectorIndexConfig"]["efConstruction"],
            flat_search_cutoff=schema["vectorIndexConfig"]["flatSearchCutoff"],
            max_connections=schema["vectorIndexConfig"]["maxConnections"],
            pq=_PQConfig(
                enabled=schema["vectorIndexConfig"]["pq"]["enabled"],
                bit_compression=schema["vectorIndexConfig"]["pq"]["bitCompression"],
                segments=schema["vectorIndexConfig"]["pq"]["segments"],
                centroids=schema["vectorIndexConfig"]["pq"]["centroids"],
                training_limit=schema["vectorIndexConfig"]["pq"]["trainingLimit"],
                encoder=_PQEncoderConfig(
                    type_=PQEncoderType(schema["vectorIndexConfig"]["pq"]["encoder"]["type"]),
                    distribution=PQEncoderDistribution(
                        schema["vectorIndexConfig"]["pq"]["encoder"]["distribution"]
                    ),
                ),
            ),
            skip=schema["vectorIndexConfig"]["skip"],
            vector_cache_max_objects=schema["vectorIndexConfig"]["vectorCacheMaxObjects"],
        ),
        vector_index_type=_VectorIndexType(schema["vectorIndexType"]),
        vectorizer=Vectorizer(schema["vectorizer"]),
    )


def _collection_configs_from_json(schema: Dict[str, Any]) -> Dict[str, _CollectionConfig]:
    return {schema["class"]: _collection_config_from_json(schema) for schema in schema["classes"]}


def _collection_configs_simple_from_json(
    schema: Dict[str, Any]
) -> Dict[str, _CollectionConfigSimple]:
    return {
        schema["class"]: _collection_config_simple_from_json(schema) for schema in schema["classes"]
    }



import math
import time
from abc import ABC, abstractmethod
from collections import deque
from concurrent.futures import Future, ThreadPoolExecutor, as_completed
from threading import Event, Thread
from typing import Any, Deque, Dict, Generic, List, Optional, Sequence, Set, Tuple, TypeVar, cast

from pydantic import ValidationError

from requests import ReadTimeout
from requests.exceptions import ConnectionError as RequestsConnectionError
from requests.exceptions import HTTPError as RequestsHTTPError

from weaviate.cluster import Cluster
from weaviate.collections.classes.batch import (
    BatchObject,
    BatchReference,
    ErrorObject,
    ErrorReference,
    _BatchObject,
    BatchObjectReturn,
    _BatchReference,
    BatchReferenceReturn,
    Shard,
)
from weaviate.collections.classes.config import ConsistencyLevel
from weaviate.collections.batch.grpc import _BatchGRPC
from weaviate.collections.batch.rest import _BatchREST
from weaviate.connect import Connection
from weaviate.exceptions import WeaviateBatchValidationError
from weaviate.util import _capitalize_first_letter, _decode_json_response_list
from weaviate.warnings import _Warnings
from weaviate.types import UUID, WeaviateField


class BatchExecutor(ThreadPoolExecutor):
    """
    Weaviate Batch Executor to run batch requests in separate thread.

    This class implements an additional method `_is_shutdown` that is used by the context manager.
    """

    def _is_shutdown(self) -> bool:
        return self._shutdown


BatchResponse = List[Dict[str, Any]]


TBatchInput = TypeVar("TBatchInput")
TBatchReturn = TypeVar("TBatchReturn")


class BatchRequest(ABC, Generic[TBatchInput, TBatchReturn]):
    """`BatchRequest` abstract class used as a interface for batch requests."""

    def __init__(self) -> None:
        self.__items: Deque[TBatchInput] = deque([])

    def __len__(self) -> int:
        return len(self.__items)

    def is_empty(self) -> bool:
        """Check if `BatchRequest` is empty.

        Returns
            `bool` Whether the `BatchRequest` is empty.
        """
        return len(self.__items) == 0

    def clear(self) -> None:
        """Remove all the items from the BatchRequest."""
        self.__items.clear()

    def add(self, item: TBatchInput) -> None:
        """Add an item to the BatchRequest."""
        self.__items.append(item)

    @property
    def items(self) -> Deque[TBatchInput]:
        """
        Get all items from the BatchRequest.

        Returns
            `Deque[TBatchInput]` All items from the BatchRequest.
        """
        return self.__items

    @abstractmethod
    def _add_failed_objects_from_response(
        self,
        response_item: TBatchReturn,
        errors_to_exclude: Optional[List[str]] = None,
        errors_to_include: Optional[List[str]] = None,
    ) -> None:
        """Add failed items from a weaviate response.

        Arguments:
            `response_item`
                Weaviate response that contains the status for all objects.
            `errors_to_exclude`
                Which errors should NOT be retried.
            `errors_to_include`
                Which errors should be retried.

        Returns:
            `BatchResponse`
                Contains responses form all successful object, eg. those that have not been added to this batch.
        """

    @staticmethod
    def _skip_objects_retry(
        entry: Dict[str, Any],
        errors_to_exclude: Optional[List[str]],
        errors_to_include: Optional[List[str]],
    ) -> bool:
        if (
            len(entry["result"]) == 0
            or "errors" not in entry["result"]
            or "error" not in entry["result"]["errors"]
            or len(entry["result"]["errors"]["error"]) == 0
        ):
            return True

        # skip based on error messages
        if errors_to_exclude is not None:
            for err in entry["result"]["errors"]["error"]:
                if any(excl in err["message"] for excl in errors_to_exclude):
                    return True
            return False
        elif errors_to_include is not None:
            for err in entry["result"]["errors"]["error"]:
                if any(incl in err["message"] for incl in errors_to_include):
                    return False
            return True
        return False


class ReferencesBatchRequest(BatchRequest[_BatchReference, BatchReferenceReturn]):
    """
    Collect Weaviate-object references to add them in one request to Weaviate.

    Caution this request will miss some validations to be faster.
    """

    def _add_failed_objects_from_response(
        self,
        return_: BatchReferenceReturn,
        errors_to_exclude: Optional[List[str]] = None,
        errors_to_include: Optional[List[str]] = None,
    ) -> None:
        # successful_responses = []

        for err in return_.errors.values():
            # if self._skip_objects_retry(ref, errors_to_exclude, errors_to_include):
            #     successful_responses.append(ref)
            #     continue
            self.add(err.reference)
        # return successful_responses


class ObjectsBatchRequest(BatchRequest[_BatchObject, BatchObjectReturn]):
    """
    Collect objects for one batch request to weaviate.

    Caution this batch will not be validated through weaviate.
    """

    def _add_failed_objects_from_response(
        self,
        return_: BatchObjectReturn,
        errors_to_exclude: Optional[List[str]] = None,
        errors_to_include: Optional[List[str]] = None,
    ) -> None:
        # successful_responses = []
        if return_.has_errors:
            for err in return_.errors.values():
                # if self._skip_objects_retry(obj, errors_to_exclude, errors_to_include):
                #     successful_responses.append(obj)
                #     continue
                self.add(err.object_)
        # return successful_responses


class _Batch:
    def __init__(
        self,
        connection: Connection,
        objects_: Optional[ObjectsBatchRequest] = None,
        references: Optional[ReferencesBatchRequest] = None,
    ) -> None:
        self.__batch_objects = objects_ or ObjectsBatchRequest()
        self.__batch_references = references or ReferencesBatchRequest()
        self.__connection = connection
        self.__consistency_level: Optional[ConsistencyLevel] = None
        self.__creation_time = min(self.__connection.timeout_config[1] / 10, 2)
        self.__batch_grpc = _BatchGRPC(connection, self.__consistency_level)
        self.__batch_rest = _BatchREST(connection, self.__consistency_level)
        self.__executor: Optional[BatchExecutor] = None
        self.__failed_objects: List[ErrorObject] = []
        self.__failed_references: List[ErrorReference] = []
        self.__future_pool_objects: Deque[Future[Tuple[BatchObjectReturn, int, bool]]] = deque([])
        self.__future_pool_references: Deque[
            Future[Tuple[BatchReferenceReturn, int, bool]]
        ] = deque([])
        self.__dynamic_batching = True
        self.__num_workers: int = 1
        self.__objects_throughput_frame: Deque[float] = deque(maxlen=5)
        self.__recommended_num_objects = 50
        self.__recommended_num_references = 50
        self.__reference_batch_queue: Deque[List[_BatchReference]] = deque([])
        self.__references_throughput_frame: Deque[float] = deque(maxlen=5)
        self.__retry_failed_objects: bool = False
        self.__retry_failed_references: bool = False
        self.__shut_background_thread_down: Optional[Event] = None
        self.__imported_shards: Set[Shard] = set()

    def configure(
        self,
        dynamic: bool = True,
        consistency_level: Optional[ConsistencyLevel] = None,
        num_workers: Optional[int] = None,
        retry_failed_objects: bool = False,
        retry_failed_references: bool = False,
    ) -> None:
        """
        Configure your batch manager.

        Every time you run this command, the `client.batch` object will
        be updated with the new configuration. To enter the batching context manager, which handles automatically
        sending batches dynamically, use `with client.batch as batch` and then loop through your data within the context manager
        adding objects and references to the batch.

        Batches are constructed automatically and sent dynamically depending on Weaviate's load.
        When you exit the context manager, the final batch will be sent automatically.

        Arguments:
            `dynamic`
                Whether to use dynamic batching or not. If not provided, the default value is True.
            `consistency_level`
                The consistency level to be used to send the batch. If not provided, the default value is None.
            `num_workers`
                The number of workers to be used when sending the batches. If not provided, the default value is 1.
                This controls the number of concurrent requests made to Weaviate and not the speed of batch creation within Python.
            `retry_failed_objects`
                Whether to retry failed objects or not. If not provided, the default value is False.
            `retry_failed_references`
                Whether to retry failed references or not. If not provided, the default value is False.
        """
        self.__consistency_level = consistency_level or self.__consistency_level
        self.__dynamic_batching = dynamic
        self.__num_workers = num_workers or self.__num_workers
        self.__retry_failed_objects = retry_failed_objects
        self.__retry_failed_references = retry_failed_references

    def num_objects(self) -> int:
        """
        Get current number of objects in the batch.

        Returns:
            `int`
                The number of objects in the batch.
        """
        return len(self.__batch_objects)

    def num_references(self) -> int:
        """
        Get current number of references in the batch.

        Returns:
            `int`
                The number of references in the batch.
        """
        return len(self.__batch_references)

    def failed_objects(self) -> List[ErrorObject]:
        """
        Get all failed objects from the batch manager.

        Returns:
            `List[ErrorObject]`
                A list of all the failed objects from the batch.
        """
        return self.__failed_objects

    def failed_references(self) -> List[ErrorReference]:
        """
        Get all failed references from the batch manager.

        Returns:
            `List[ErrorReference]`
                A list of all the failed references from the batch.
        """
        return self.__failed_references

    def start(self) -> "_Batch":
        """
        Start the BatchExecutor if it was closed.

        Returns:
            `Batch`
                The batch object with an open BatchExecutor and background running thread.
        """
        if self.__executor is None or self.__executor._is_shutdown():
            self.__executor = BatchExecutor(max_workers=self.__num_workers)

        if (
            self.__shut_background_thread_down is None
            or self.__shut_background_thread_down.is_set()
        ):
            self.__update_recommended_batch_size()

        return self

    def shutdown(self) -> None:
        """Shutdown the BatchExecutor."""
        if not (self.__executor is None or self.__executor._is_shutdown()):
            self.__executor.shutdown()

        if self.__shut_background_thread_down is not None:
            self.__shut_background_thread_down.set()

    def flush(self) -> None:
        """Flush the batch queue to ensure all batches are sent."""
        self.__send_batch_requests(force_wait=True)

    def __enter__(self) -> "_Batch":
        if not self.__dynamic_batching:
            raise WeaviateBatchValidationError(
                "Cannot use the batch context manager without dynamic batching. If you are doing manual batching, you must create them manually outside of the context manager with create_objects()"
            )
        self.__reset()
        return self.start()

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        self.flush()
        self.shutdown()

    def __reset(self) -> None:
        self.__batch_objects.clear()
        self.__batch_references.clear()
        self.__future_pool_objects.clear()
        self.__future_pool_references.clear()
        self.__failed_objects.clear()
        self.__failed_references.clear()
        self.__reference_batch_queue.clear()

    def add_object(
        self,
        class_name: str,
        properties: Dict[str, WeaviateField],
        uuid: Optional[UUID] = None,
        vector: Optional[Sequence] = None,
        tenant: Optional[str] = None,
    ) -> UUID:
        """
        Add one object to this batch.

        NOTE: If the UUID of one of the objects already exists then the existing object will be
        replaced by the new object.

        Arguments:
            `class_name`
                The name of the class this object belongs to.
            `properties`
                The data properties of the object to be added as a dictionary.
            `uuid`:
                The UUID of the object as an uuid.UUID object or str. It can be a Weaviate beacon or Weaviate href.
                If it is None an UUIDv4 will generated, by default None
            `vector`:
                The embedding of the object that should be validated. Can be used when a class does not have a vectorization module or the given vector was generated using the _identical_ vectorization module that is configured for the class. In this case this vector takes precedence. Supported types are `list`, 'numpy.ndarray`, `torch.Tensor` and `tf.Tensor`, by default None.
            `tenant`
                Name of the tenant.

        Returns:
            `str`
                The UUID of the added object. If one was not provided a UUIDv4 will be auto-generated for you and returned here.

        Raises:
            `WeaviateBatchValidationError`
                If the provided options are in the format required by Weaviate.
        """
        try:
            batch_object = BatchObject(
                class_name=class_name,
                properties=properties,
                uuid=uuid,
                vector=vector,
                tenant=tenant,
            )
            self.__imported_shards.add(Shard(class_name=class_name, tenant=tenant))
        except ValidationError as e:
            raise WeaviateBatchValidationError(repr(e))
        self.__batch_objects.add(batch_object._to_internal())
        self.__auto_create()
        assert batch_object.uuid is not None
        return batch_object.uuid

    def add_reference(
        self,
        from_object_uuid: UUID,
        from_object_class_name: str,
        from_property_name: str,
        to_object_uuid: UUID,
        to_object_class_name: Optional[str] = None,
        tenant: Optional[str] = None,
    ) -> None:
        """
        Add one reference to this batch.

        Arguments:
            `from_object_uuid`
                The UUID of the object, as an uuid.UUID object or str, that should reference another object.
                It can be a Weaviate beacon or Weaviate href.
            `from_object_class_name`
                The name of the class that should reference another object.
            `from_property_name`
                The name of the property that contains the reference.
            `to_object_uuid`
                The UUID of the object, as an uuid.UUID object or str, that is actually referenced.
                It can be a Weaviate beacon or Weaviate href.
            `to_object_class_name`
                The referenced object class name to which to add the reference (with UUID `to_object_uuid`).
            `tenant`
                Name of the tenant.

        Raises:
            `WeaviateBatchValidationError`
                If the provided options are in the format required by Weaviate.
        """
        try:
            batch_reference = BatchReference(
                from_object_class_name=from_object_class_name,
                from_object_uuid=from_object_uuid,
                from_property_name=from_property_name,
                to_object_class_name=to_object_class_name,
                to_object_uuid=to_object_uuid,
                tenant=tenant,
            )
        except ValidationError as e:
            raise WeaviateBatchValidationError(repr(e))
        self.__batch_references.add(batch_reference._to_internal())
        self.__auto_create()

    def create_objects(self) -> BatchObjectReturn:
        """
        Send all the currently batched objects to Weaviate manually.

        This function will do nothing unless you set `dynamic` to `False` in `configure`.
        """
        if self.__dynamic_batching:
            _Warnings.batch_create_dynamic("objects")
            return BatchObjectReturn(
                all_responses=[],
                uuids={},
                errors={},
                elapsed_seconds=0,
                has_errors=False,
            )
        res = self.__flush_objects(list(self.__batch_objects.items))
        self.__batch_objects.clear()
        return res[0]

    def create_references(self) -> BatchReferenceReturn:
        """
        Send all the currently batched references to Weaviate manually.

        This function will do nothing unless you set `dynamic` to `False` in `configure`.
        """
        if self.__dynamic_batching:
            _Warnings.batch_create_dynamic("references")
            return BatchReferenceReturn(
                elapsed_seconds=0,
                errors={},
                has_errors=False,
            )
        res = self.__flush_references(list(self.__batch_references.items))
        self.__batch_references.clear()
        return res[0]

    def __auto_create(self) -> None:
        if (
            self.num_objects() >= self.__recommended_num_objects
            or self.num_references() >= self.__recommended_num_references
        ):
            while self.__recommended_num_objects == 0 or self.__recommended_num_references == 0:
                sleep = 1
                _Warnings.batch_weaviate_overloaded_sleeping(sleep)
                time.sleep(sleep)  # block if weaviate is overloaded
            self.__send_batch_requests(force_wait=False)

    def __flush_objects(self, batch: List[_BatchObject]) -> Tuple[BatchObjectReturn, int, bool]:
        start = time.time()
        try:
            return_ = self.__batch_grpc.objects(
                objects=batch,
            )
            return return_, len(batch), False
        except Exception as e:
            print(repr(e), batch)
            errors = {
                idx: ErrorObject(message=repr(e), object_=obj) for idx, obj in enumerate(batch)
            }
            return (
                BatchObjectReturn(
                    all_responses=list(errors.values()),
                    elapsed_seconds=time.time() - start,
                    errors=errors,
                    has_errors=True,
                    uuids={},
                ),
                0,
                True,
            )

    def __flush_references(
        self, batch: List[_BatchReference]
    ) -> Tuple[BatchReferenceReturn, int, bool]:
        start = time.time()
        try:
            response = self.__batch_rest.references(
                references=batch,
            )
            return response, len(batch), False
        except Exception as e:
            print(repr(e), batch)
            errors = {
                idx: ErrorReference(message=repr(e), reference=ref) for idx, ref in enumerate(batch)
            }
            return (
                BatchReferenceReturn(
                    elapsed_seconds=time.time() - start,
                    errors=errors,
                    has_errors=True,
                ),
                0,
                True,
            )

    def __send_batch_requests(self, force_wait: bool, how_many_recursions: int = 0) -> None:
        if self.__executor is None:
            self.start()
        elif self.__executor._is_shutdown():
            _Warnings.batch_executor_is_shutdown()
            self.start()

        assert self.__executor is not None

        if len(self.__batch_objects) > 0:
            self.__future_pool_objects.append(
                self.__executor.submit(
                    self.__flush_objects,
                    batch=list(self.__batch_objects.items),
                )
            )
        if len(self.__batch_references) > 0:
            # convert deque to list to ensure data is copied before being cleared below
            self.__reference_batch_queue.append(list(self.__batch_references.items))

        self.__batch_objects.clear()
        self.__batch_references.clear()

        if (
            not force_wait
            and self.__num_workers > 1
            and len(self.__future_pool_objects) < self.__num_workers
        ):
            return

        for done_obj_future in as_completed(self.__future_pool_objects):
            ret_objs, nr_objs, exception_raised = done_obj_future.result()
            self.__failed_objects.extend(ret_objs.errors.values())
            if self.__retry_failed_objects and (exception_raised or ret_objs.has_errors):
                self.__batch_objects._add_failed_objects_from_response(ret_objs)
                self.__backoff_recommended_object_batch_size(exception_raised)
            else:
                self.__objects_throughput_frame.append(nr_objs / ret_objs.elapsed_seconds)

        for ref_batch_items in self.__reference_batch_queue:
            self.__future_pool_references.append(
                self.__executor.submit(
                    self.__flush_references,
                    batch=list(ref_batch_items),
                )
            )

        for done_ref_future in as_completed(self.__future_pool_references):
            ret_refs, nr_refs, exception_raised = done_ref_future.result()
            self.__failed_references.extend(ret_refs.errors.values())
            if self.__retry_failed_references and (exception_raised or ret_refs.has_errors):
                self.__batch_references._add_failed_objects_from_response(ret_refs)
                self.__backoff_recommended_reference_batch_size(exception_raised)
            else:
                self.__references_throughput_frame.append(nr_refs / ret_refs.elapsed_seconds)

        # Clear futures before checking if we need to retry
        self.__future_pool_objects.clear()
        self.__future_pool_references.clear()
        self.__reference_batch_queue.clear()

        if len(self.__batch_objects) > 0 or len(self.__batch_references) > 0:
            if how_many_recursions == 4:
                _Warnings.batch_retrying_failed_batches_hit_hard_limit(5)
            else:
                self.__failed_objects.clear()
                self.__failed_references.clear()
                self.__send_batch_requests(
                    force_wait=True, how_many_recursions=how_many_recursions + 1
                )

    def __backoff_recommended_object_batch_size(self, exception_occurred: bool) -> None:
        if exception_occurred and self.__recommended_num_objects is not None:
            self.__recommended_num_objects = max(self.__recommended_num_objects // 2, 1)
        elif (
            len(self.__objects_throughput_frame) != 0 and self.__recommended_num_objects is not None
        ):
            obj_per_second = (
                sum(self.__objects_throughput_frame) / len(self.__objects_throughput_frame) * 0.75
            )
            self.__recommended_num_objects = max(
                min(
                    round(obj_per_second * self.__creation_time),
                    self.__recommended_num_objects + 250,
                ),
                1,
            )

    def __backoff_recommended_reference_batch_size(self, exception_occurred: bool) -> None:
        if exception_occurred and self.__recommended_num_references is not None:
            self.__recommended_num_references = max(self.__recommended_num_references // 2, 1)
        elif (
            len(self.__references_throughput_frame) != 0
            and self.__recommended_num_references is not None
        ):
            ref_per_sec = sum(self.__references_throughput_frame) / len(
                self.__references_throughput_frame
            )
            self.__recommended_num_references = min(
                round(ref_per_sec * self.__creation_time),
                self.__recommended_num_references * 2,
            )

    def __update_recommended_batch_size(self) -> None:
        """Create a background process that periodically checks how congested the batch queue is."""
        self.__shut_background_thread_down = Event()

        def periodic_check() -> None:
            cluster = Cluster(self.__connection)
            while (
                self.__shut_background_thread_down is not None
                and not self.__shut_background_thread_down.is_set()
            ):
                try:
                    status = cluster.get_nodes_status()
                    if "stats" not in status[0] or "ratePerSecond" not in status[0]["stats"]:
                        self.__dynamic_batching = False
                        return
                    rate = status[0]["batchStats"]["ratePerSecond"]
                    rate_per_worker = rate / self.__num_workers
                    batch_length = status[0]["batchStats"]["queueLength"]

                    if batch_length == 0:  # scale up if queue is empty
                        self.__recommended_num_objects = self.__recommended_num_objects + min(
                            self.__recommended_num_objects * 2, 25
                        )
                    else:
                        ratio = batch_length / rate
                        if (
                            2.1 > ratio > 1.9
                        ):  # ideal, send exactly as many objects as weaviate can process
                            self.__recommended_num_objects = math.floor(rate_per_worker)
                        elif ratio <= 1.9:  # we can send more
                            self.__recommended_num_objects = math.floor(
                                min(
                                    self.__recommended_num_objects * 1.5,
                                    rate_per_worker * 2 / ratio,
                                )
                            )
                        elif ratio < 10:  # too high, scale down
                            self.__recommended_num_objects = math.floor(rate_per_worker * 2 / ratio)
                        else:  # way too high, stop sending new batches
                            self.__recommended_num_objects = 0

                    refresh_time: float = 2
                except (RequestsHTTPError, ReadTimeout):
                    refresh_time = 0.1
                except Exception as e:
                    _Warnings.batch_refresh_failed(repr(e))
                    refresh_time = 10

                time.sleep(refresh_time)
            self.__recommended_num_objects = 10  # in case some batch needs to be send afterwards
            self.__shut_background_thread_down = None

        demon = Thread(
            target=periodic_check,
            daemon=True,
            name="batchSizeRefresh",
        )
        demon.start()

    def wait_for_vector_indexing(
        self, shards: Optional[List[Shard]] = None, how_many_failures: int = 5
    ) -> None:
        """Wait for the all the vectors of the batch imported objects to be indexed.

        Upon network error, it will retry to get the shards' status for `how_many_failures` times
        with exponential backoff (2**n seconds with n=0,1,2,...,how_many_failures).

        Parameters
        ----------
            shards {Optional[List[Shard]]} -- The shards to check the status of. If None it will
                check the status of all the shards of the imported objects in the batch.
            how_many_failures {int} -- How many times to try to get the shards' status before
                raising an exception. Default 5.
        """
        if shards is not None and not isinstance(shards, list):
            raise TypeError(f"'shards' must be of type List[Shard]. Given type: {type(shards)}.")
        if shards is not None and not isinstance(shards[0], Shard):
            raise TypeError(f"'shards' must be of type List[Shard]. Given type: {type(shards)}.")

        def is_ready(how_many: int) -> bool:
            try:
                return all(
                    all(self._get_shards_readiness(shard))
                    for shard in shards or self.__imported_shards
                )
            except RequestsConnectionError as e:
                print(
                    f"Error while getting class shards statuses: {e}, trying again with 2**n={2**how_many}s exponential backoff with n={how_many}"
                )
                if how_many_failures == how_many:
                    raise e
                time.sleep(2**how_many)
                return is_ready(how_many + 1)

        while not is_ready(0):
            print("Waiting for async indexing to finish...")
            time.sleep(0.25)

    def _get_shards_readiness(self, shard: Shard) -> List[bool]:
        if not isinstance(shard.class_name, str):
            raise TypeError(
                "'class_name' argument must be of type `str`! "
                f"Given type: {type(shard.class_name)}."
            )

        path = f"/schema/{_capitalize_first_letter(shard.class_name)}/shards{'' if shard.tenant is None else f'?tenant={shard.tenant}'}"

        try:
            response = self.__connection.get(path=path)
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError(
                "Class shards' status could not be retrieved due to connection error."
            ) from conn_err

        res = _decode_json_response_list(response, "Get shards' status")
        assert res is not None
        return [
            (cast(str, shard.get("status")) == "READY")
            & (cast(int, shard.get("vectorQueueSize")) == 0)
            for shard in res
        ]



import pathlib

import pytest

import weaviate
from weaviate.collections.classes.aggregate import Metrics
from weaviate.collections.classes.config import DataType, Property, ReferenceProperty, Configure
from weaviate.exceptions import WeaviateInvalidInputException
from weaviate.util import file_encoder_b64


@pytest.fixture(scope="module")
def client():
    client = weaviate.connect_to_local()
    client.collections.delete_all()
    yield client
    client.collections.delete_all()


@pytest.mark.parametrize("how_many", [1, 10000, 20000, 20001, 100000])
def test_collection_length(client: weaviate.WeaviateClient, how_many: int):
    """Uses .aggregate behind-the-scenes"""
    name = "TestCollectionLength"
    client.collections.delete(name)
    collection = client.collections.create(
        name=name,
        properties=[Property(name="Name", data_type=DataType.TEXT)],
        vectorizer_config=Configure.Vectorizer.none(),
    )
    collection.data.insert_many([{"Name": f"name {i}"} for i in range(how_many)])
    assert len(collection) == how_many


def test_simple_aggregation(client: weaviate.WeaviateClient):
    name = "TestSimpleAggregation"
    client.collections.delete(name)
    collection = client.collections.create(
        name=name, properties=[Property(name="text", data_type=DataType.TEXT)]
    )
    collection.data.insert({"text": "some text"})
    res = collection.aggregate.over_all(
        return_metrics=[Metrics("text", DataType.TEXT).returning(count=True)]
    )
    assert res.properties["text"].count == 1


@pytest.mark.parametrize(
    "option,expected_len",
    [
        ({"object_limit": 1}, 1),
        ({"certainty": 0.9}, 1),
        ({"distance": 0.1}, 1),
        ({"object_limit": 2}, 2),
        ({"certainty": 0.1}, 2),
        ({"distance": 0.9}, 2),
    ],
)
def test_near_object_aggregation(client: weaviate.WeaviateClient, option: dict, expected_len: int):
    name = "TestNearObjectAggregation"
    client.collections.delete(name)
    collection = client.collections.create(
        name=name,
        properties=[Property(name="text", data_type=DataType.TEXT)],
        vectorizer_config=Configure.Vectorizer.text2vec_contextionary(vectorize_class_name=False),
    )
    text_1 = "some text"
    text_2 = "nothing like the other one at all, not even a little bit"
    uuid = collection.data.insert({"text": text_1})
    collection.data.insert({"text": text_2})
    res = collection.aggregate.near_object(
        uuid,
        return_metrics=[
            Metrics("text", DataType.TEXT).returning(
                count=True, top_occurrences_count=True, top_occurrences_value=True
            )
        ],
        **option,
    )
    assert res.properties["text"].count == expected_len
    assert len(res.properties["text"].top_occurrences) == expected_len
    assert text_1 in [
        top_occurrence.value for top_occurrence in res.properties["text"].top_occurrences
    ]
    if expected_len == 2:
        assert text_2 in [
            top_occurrence.value for top_occurrence in res.properties["text"].top_occurrences
        ]
    else:
        assert text_2 not in [
            top_occurrence.value for top_occurrence in res.properties["text"].top_occurrences
        ]


def test_near_object_missing_param(client: weaviate.WeaviateClient):
    name = "TestNearVectorMissingParam"
    client.collections.delete(name)
    collection = client.collections.create(
        name=name,
        properties=[Property(name="text", data_type=DataType.TEXT)],
        vectorizer_config=Configure.Vectorizer.text2vec_contextionary(vectorize_class_name=False),
    )
    text_1 = "some text"
    text_2 = "nothing like the other one at all, not even a little bit"
    uuid = collection.data.insert({"text": text_1})
    collection.data.insert({"text": text_2})
    with pytest.raises(WeaviateInvalidInputException) as e:
        collection.aggregate.near_object(
            uuid,
            return_metrics=[
                Metrics("text", DataType.TEXT).returning(
                    count=True, top_occurrences_count=True, top_occurrences_value=True
                )
            ],
        )
    assert (
        "You must provide at least one of the following arguments: certainty, distance, object_limit when vector searching"
        == e.value.message
    )


@pytest.mark.parametrize(
    "option,expected_len",
    [
        ({"object_limit": 1}, 1),
        ({"certainty": 0.9}, 1),
        ({"distance": 0.1}, 1),
        ({"object_limit": 2}, 2),
        ({"certainty": 0.1}, 2),
        ({"distance": 0.9}, 2),
    ],
)
def test_near_vector_aggregation(client: weaviate.WeaviateClient, option: dict, expected_len: int):
    name = "TestNearVectorAggregation"
    client.collections.delete(name)
    collection = client.collections.create(
        name=name,
        properties=[Property(name="text", data_type=DataType.TEXT)],
        vectorizer_config=Configure.Vectorizer.text2vec_contextionary(vectorize_class_name=False),
    )
    text_1 = "some text"
    text_2 = "nothing like the other one at all, not even a little bit"
    uuid = collection.data.insert({"text": text_1})
    obj = collection.query.fetch_object_by_id(uuid, include_vector=True)
    collection.data.insert({"text": text_2})
    res = collection.aggregate.near_vector(
        obj.metadata.vector,
        return_metrics=[
            Metrics("text", DataType.TEXT).returning(
                count=True, top_occurrences_count=True, top_occurrences_value=True
            )
        ],
        **option,
    )
    assert res.properties["text"].count == expected_len
    assert len(res.properties["text"].top_occurrences) == expected_len
    assert text_1 in [
        top_occurrence.value for top_occurrence in res.properties["text"].top_occurrences
    ]
    if expected_len == 2:
        assert text_2 in [
            top_occurrence.value for top_occurrence in res.properties["text"].top_occurrences
        ]
    else:
        assert text_2 not in [
            top_occurrence.value for top_occurrence in res.properties["text"].top_occurrences
        ]


def test_near_vector_missing_param(client: weaviate.WeaviateClient):
    name = "TestNearVectorMissingParam"
    client.collections.delete(name)
    collection = client.collections.create(
        name=name,
        properties=[Property(name="text", data_type=DataType.TEXT)],
        vectorizer_config=Configure.Vectorizer.text2vec_contextionary(vectorize_class_name=False),
    )
    uuid_ = collection.data.insert({"text": "some text"})
    obj = collection.query.fetch_object_by_id(uuid_, include_vector=True)
    with pytest.raises(WeaviateInvalidInputException) as e:
        collection.aggregate.near_vector(
            obj.metadata.vector,
            return_metrics=[
                Metrics("text", DataType.TEXT).returning(
                    count=True, top_occurrences_count=True, top_occurrences_value=True
                )
            ],
        )
    assert (
        "You must provide at least one of the following arguments: certainty, distance, object_limit when vector searching"
        == e.value.message
    )


@pytest.mark.parametrize(
    "option,expected_len",
    [
        ({"object_limit": 1}, 1),
        ({"certainty": 0.9}, 1),
        ({"distance": 0.1}, 1),
        ({"object_limit": 2}, 2),
        ({"certainty": 0.1}, 2),
        ({"distance": 0.9}, 2),
    ],
)
def test_near_text_aggregation(client: weaviate.WeaviateClient, option: dict, expected_len: int):
    name = "TestNearTextAggregation"
    client.collections.delete(name)
    collection = client.collections.create(
        name=name,
        properties=[Property(name="text", data_type=DataType.TEXT)],
        vectorizer_config=Configure.Vectorizer.text2vec_contextionary(vectorize_class_name=False),
    )
    text_1 = "some text"
    text_2 = "nothing like the other one at all, not even a little bit"
    collection.data.insert({"text": text_1})
    collection.data.insert({"text": text_2})
    res = collection.aggregate.near_text(
        text_1,
        return_metrics=[
            Metrics("text", DataType.TEXT).returning(
                count=True, top_occurrences_count=True, top_occurrences_value=True
            )
        ],
        **option,
    )
    assert res.properties["text"].count == expected_len
    assert len(res.properties["text"].top_occurrences) == expected_len
    assert text_1 in [
        top_occurrence.value for top_occurrence in res.properties["text"].top_occurrences
    ]
    if expected_len == 2:
        assert text_2 in [
            top_occurrence.value for top_occurrence in res.properties["text"].top_occurrences
        ]
    else:
        assert text_2 not in [
            top_occurrence.value for top_occurrence in res.properties["text"].top_occurrences
        ]


def test_near_text_missing_param(client: weaviate.WeaviateClient):
    name = "TestNearTextMissingParam"
    client.collections.delete(name)
    collection = client.collections.create(
        name=name,
        properties=[Property(name="text", data_type=DataType.TEXT)],
        vectorizer_config=Configure.Vectorizer.text2vec_contextionary(vectorize_class_name=False),
    )
    text_1 = "some text"
    collection.data.insert({"text": text_1})
    with pytest.raises(WeaviateInvalidInputException) as e:
        collection.aggregate.near_text(
            text_1,
            return_metrics=[
                Metrics("text", DataType.TEXT).returning(
                    count=True, top_occurrences_count=True, top_occurrences_value=True
                )
            ],
        )
    assert (
        "You must provide at least one of the following arguments: certainty, distance, object_limit when vector searching"
        == e.value.message
    )


@pytest.mark.parametrize("option", [{"object_limit": 1}, {"certainty": 0.9}, {"distance": 0.1}])
def test_near_image_aggregation(client: weaviate.WeaviateClient, option: dict):
    name = "TestNearImageAggregation"
    client.collections.delete(name)
    collection = client.collections.create(
        name=name,
        properties=[
            Property(name="rating", data_type=DataType.INT),
            Property(name="image", data_type=DataType.BLOB),
        ],
        vectorizer_config=Configure.Vectorizer.img2vec_neural(image_fields=["image"]),
    )
    img_path = pathlib.Path("integration/weaviate-logo.png")
    collection.data.insert({"image": file_encoder_b64(img_path), "rating": 9})
    res = collection.aggregate.near_image(
        img_path,
        return_metrics=[Metrics("rating", DataType.INT).returning(count=True, maximum=True)],
        **option,
    )
    assert res.properties["rating"].count == 1
    assert res.properties["rating"].maximum == 9


def test_near_image_missing_param(client: weaviate.WeaviateClient):
    name = "TestNearImageMissingParam"
    client.collections.delete(name)
    collection = client.collections.create(
        name=name,
        properties=[
            Property(name="rating", data_type=DataType.INT),
            Property(name="image", data_type=DataType.BLOB),
        ],
        vectorizer_config=Configure.Vectorizer.img2vec_neural(image_fields=["image"]),
    )
    img_path = pathlib.Path("integration/weaviate-logo.png")
    collection.data.insert({"image": file_encoder_b64(img_path), "rating": 9})
    with pytest.raises(WeaviateInvalidInputException) as e:
        collection.aggregate.near_image(
            img_path,
            return_metrics=[
                Metrics("text", DataType.TEXT).returning(
                    count=True, top_occurrences_count=True, top_occurrences_value=True
                )
            ],
        )
    assert (
        "You must provide at least one of the following arguments: certainty, distance, object_limit when vector searching"
        == e.value.message
    )


def test_group_by_aggregation(client: weaviate.WeaviateClient):
    name = "TestGroupByAggregation"
    client.collections.delete(name)
    collection = client.collections.create(
        name=name,
        properties=[
            Property(name="text", data_type=DataType.TEXT),
            Property(name="int", data_type=DataType.INT),
        ],
    )
    collection.data.insert({"text": "some text", "int": 1})
    collection.data.insert({"text": "some text", "int": 2})

    res = collection.aggregate_group_by.over_all(
        "text",
        return_metrics=[
            Metrics("text", DataType.TEXT).returning(count=True),
            Metrics("int", DataType.INT).returning(count=True),
        ],
    )
    assert len(res) == 1
    assert res[0].grouped_by.prop == "text"
    assert res[0].grouped_by.value == "some text"
    assert res[0].properties["text"].count == 2
    assert res[0].properties["int"].count == 2

    res = collection.aggregate_group_by.over_all(
        "int",
        return_metrics=[
            Metrics("text", DataType.TEXT).returning(count=True),
            Metrics("int", DataType.INT).returning(count=True),
        ],
    )
    assert len(res) == 2
    assert res[0].grouped_by.prop == "int"
    assert res[0].grouped_by.value == "1" or res[1].grouped_by.value == "1"
    assert res[0].properties["text"].count == 1
    assert res[0].properties["int"].count == 1
    assert res[1].grouped_by.prop == "int"
    assert res[1].grouped_by.value == "2" or res[0].grouped_by.value == "2"
    assert res[1].properties["text"].count == 1
    assert res[1].properties["int"].count == 1


@pytest.mark.skip(reason="Validation logic is not robust enough currently")
def test_mistake_in_usage(client: weaviate.WeaviateClient):
    collection = client.collections.get("TestMistakeInUsage")
    with pytest.raises(TypeError) as e:
        collection.aggregate.over_all([Metrics("text", DataType.TEXT)])
    assert (
        e.value.args[0]
        == "One of the aggregations is an unexpected type: <class 'weaviate.collection.classes.aggregate.Metrics'>. Did you forget to append .returning() to .with_()?"
    )
    with pytest.raises(TypeError) as e:
        collection.aggregate.over_all(aggregations=[Metrics("text", DataType.TEXT)])
    assert (
        e.value.args[0]
        == "One of the aggregations is an unexpected type: <class 'weaviate.collection.classes.aggregate.Metrics'>. Did you forget to append .returning() to .with_()?"
    )


def test_all_available_aggregations(client: weaviate.WeaviateClient):
    name = "TestAllAvailableAggregations"
    client.collections.delete(name)
    collection = client.collections.create(
        name=name,
        properties=[
            Property(name="text", data_type=DataType.TEXT),
            Property(
                name="texts",
                data_type=DataType.TEXT_ARRAY,
            ),
            Property(name="int", data_type=DataType.INT),
            Property(name="ints", data_type=DataType.INT_ARRAY),
            Property(name="float", data_type=DataType.NUMBER),
            Property(name="floats", data_type=DataType.NUMBER_ARRAY),
            Property(name="bool", data_type=DataType.BOOL),
            Property(name="bools", data_type=DataType.BOOL_ARRAY),
            Property(name="date", data_type=DataType.DATE),
            Property(name="dates", data_type=DataType.DATE_ARRAY),
            ReferenceProperty(name="ref", target_collection="TestAllAvailableAggregations"),
        ],
    )
    collection.data.insert(
        {
            "text": "some text",
            "texts": ["some text", "some more text"],
            "int": 1,
            "ints": [1, 2],
            "float": 1.0,
            "floats": [1.0, 2.0],
            "bool": True,
            "bools": [True, False],
            "date": "2021-01-01T00:00:00Z",
            "dates": ["2021-01-01T00:00:00Z", "2021-01-02T00:00:00Z"],
        }
    )
    res = collection.aggregate.over_all(
        return_metrics=[
            Metrics("text", DataType.TEXT).returning(count=True),
            Metrics("texts", DataType.TEXT_ARRAY).returning(count=True),
            Metrics("int", DataType.INT).returning(count=True),
            Metrics("ints", DataType.INT_ARRAY).returning(count=True),
            Metrics("float", DataType.NUMBER).returning(count=True),
            Metrics("floats", DataType.NUMBER_ARRAY).returning(count=True),
            Metrics("bool", DataType.BOOL).returning(count=True),
            Metrics("bools", DataType.BOOL_ARRAY).returning(count=True),
            Metrics("date", DataType.DATE).returning(count=True),
            Metrics("dates", DataType.DATE_ARRAY).returning(count=True),
        ]
    )
    assert res.properties["text"].count == 1
    assert res.properties["texts"].count == 2
    assert res.properties["int"].count == 1
    assert res.properties["ints"].count == 2
    assert res.properties["float"].count == 1
    assert res.properties["floats"].count == 2
    assert res.properties["bool"].count == 1
    assert res.properties["bools"].count == 2
    assert res.properties["date"].count == 1
    assert res.properties["dates"].count == 2



from typing import Any, Dict, Optional

import pytest as pytest

import weaviate

CLASS1 = {
    "class": "Test",
    "properties": [
        {"name": "test", "dataType": ["string"]},
        {"name": "abc", "dataType": ["int"]},
    ],
}

CLASS2 = {
    "class": "Test2",
    "properties": [
        {"name": "test", "dataType": ["string"]},
        {"name": "abc", "dataType": ["int"]},
        {"name": "ref", "dataType": ["Test"]},
    ],
}
VECTOR = [1.5, 2.5, 3.5] * 100  # match with vectorizer vector length


UUID1 = "577887c1-4c6b-5594-aa62-f0c17883d9bf"
UUID2 = "577887c1-4c6b-5594-aa62-f0c17883d9cf"


@pytest.mark.parametrize("grpc_port", [50051, None])
@pytest.mark.parametrize("with_limit", [True, False])
@pytest.mark.parametrize("additional_props", [None, "id", ["id"], ["id", "vector"]])
@pytest.mark.parametrize(
    "search",
    [
        {"vector": VECTOR, "certainty": 0.5},
        {"vector": VECTOR, "distance": 0.5},
        {"vector": VECTOR},
        {"id": UUID2},
        {"id": UUID2, "certainty": 0.5},
        {"id": UUID2, "distance": 0.5},
        {"bm25": ""},
        {"hybrid": ""},
    ],
)
@pytest.mark.parametrize(
    "properties",
    [
        "test",
        ["test", "abc"],
        ["test", "ref {... on Test {test abc _additional{id vector}}}"],
    ],
)
def test_grcp(
    with_limit: bool, additional_props, search: Dict[str, Any], properties, grpc_port: Optional[int]
):
    client = weaviate.Client(
        "http://localhost:8080", additional_config=weaviate.Config(grpc_port_experimental=grpc_port)
    )
    client.schema.delete_all()

    client.schema.create_class(CLASS1)
    client.schema.create_class(CLASS2)

    # add objects and references
    client.data_object.create({"test": "test"}, "Test", vector=VECTOR)
    client.data_object.create({"test": "test", "abc": 5}, "Test", vector=VECTOR, uuid=UUID1)
    client.data_object.create({"test": "test", "abc": 5}, "Test2", vector=VECTOR, uuid=UUID2)
    client.data_object.reference.add(
        from_uuid=UUID2,
        to_uuid=UUID1,
        from_class_name="Test2",
        to_class_name="Test",
        from_property_name="ref",
    )

    query = client.query.get("Test2", properties)

    if with_limit:
        query.with_limit(10)

    if additional_props is not None:
        query.with_additional(additional_props)

    if "vector" in search:
        query.with_near_vector(search)
    elif "id" in search:
        query.with_near_object(search)
    elif "concepts" in search:
        query.with_near_text(search)
    elif "bm25" in search:
        query.with_bm25(query="test", properties=["test"])
    elif "hybrid" in search:
        query.with_hybrid(query="test", properties=["test"], alpha=0.5, vector=VECTOR)

    result = query.do()
    assert "Test2" in result["data"]["Get"]
    assert "test" in result["data"]["Get"]["Test2"][0]


def test_additional():
    client_grpc = weaviate.Client(
        "http://localhost:8080", additional_config=weaviate.Config(grpc_port_experimental=50051)
    )
    client_grpc.schema.delete_all()

    client_grpc.schema.create_class(CLASS1)
    client_grpc.data_object.create({"test": "test"}, "Test", vector=VECTOR)
    client_gql = weaviate.Client(
        "http://localhost:8080", additional_config=weaviate.Config(grpc_port_experimental=50052)
    )

    results = []
    for client in [client_gql, client_grpc]:
        query = client.query.get("Test").with_additional(
            weaviate.AdditionalProperties(
                uuid=True,
                vector=True,
                creationTimeUnix=True,
                lastUpdateTimeUnix=True,
                distance=True,
            )
        )
        result = query.do()
        assert "Test" in result["data"]["Get"]

        results.append(result)

    result_gql = results[0]["data"]["Get"]["Test"][0]["_additional"]
    result_grpc = results[1]["data"]["Get"]["Test"][0]["_additional"]

    assert sorted(result_gql.keys()) == sorted(result_grpc.keys())
    for key in result_gql.keys():
        assert result_gql[key] == result_grpc[key]


def test_grpc_errors():
    client = weaviate.Client(
        "http://localhost:8080", additional_config=weaviate.Config(grpc_port_experimental=50051)
    )
    classname = CLASS1["class"]
    if client.schema.exists(classname):
        client.schema.delete_class(classname)
    client.schema.create_class(CLASS1)

    client.data_object.create({"test": "test"}, classname)

    # class errors
    res = client.query.get(classname + "does_not_exist", ["test"]).do()
    assert "errors" in res
    assert "data" not in res

    # prop errors
    res = client.query.get(classname, ["test", "made_up_prop"]).do()
    assert "errors" in res
    assert "data" not in res



import json
import os
import uuid
from typing import Optional, List, Union

import pytest
from pytest import FixtureRequest

import weaviate
from weaviate import Tenant
from weaviate.data.replication import ConsistencyLevel
from weaviate.gql.get import HybridFusion

schema = {
    "classes": [
        {
            "class": "Ship",
            "description": "object",
            "properties": [
                {"dataType": ["string"], "description": "name", "name": "name"},
                {"dataType": ["string"], "description": "description", "name": "description"},
                {"dataType": ["int"], "description": "size", "name": "size"},
                {"dataType": ["number"], "description": "rating", "name": "rating"},
            ],
            "vectorizer": "text2vec-contextionary",
        }
    ]
}

SHIPS = [
    {
        "props": {
            "name": "HMS British Name",
            "size": 5,
            "rating": 0.0,
            "description": "Super long description",
        },
        "id": uuid.uuid4(),
    },
    {
        "props": {
            "name": "The dragon ship",
            "rating": 6.66,
            "size": 20,
            "description": "Interesting things about dragons",
        },
        "id": uuid.uuid4(),
    },
    {
        "props": {
            "name": "Blackbeard",
            "size": 43,
            "rating": 7.2,
            "description": "Background info about movies",
        },
        "id": uuid.uuid4(),
    },
    {
        "props": {"name": "Titanic", "size": 1, "rating": 4.5, "description": "Everyone knows"},
        "id": uuid.uuid4(),
    },
    {
        "props": {
            "name": "Artemis",
            "size": 34,
            "rating": 9.1,
            "description": "Name from some story",
        },
        "id": uuid.uuid4(),
    },
    {
        "props": {
            "name": "The Crusty Crab",
            "size": 303,
            "rating": 10.0,
            "description": "sponges, sponges, sponges",
        },
        "id": uuid.uuid4(),
    },
]


@pytest.fixture(scope="function")
def people_schema() -> str:
    with open(os.path.join(os.path.dirname(__file__), "people_schema.json"), encoding="utf-8") as f:
        return json.load(f)


@pytest.fixture(scope="module")
def client(request):
    port = 8080
    opts = parse_client_options(request)
    if opts:
        if opts.get("cluster"):
            port = 8087
            for _, c in enumerate(schema["classes"]):
                c["replicationConfig"] = {"factor": 2}

    client = weaviate.Client(f"http://localhost:{port}")
    client.schema.delete_all()
    client.schema.create(schema)
    with client.batch as batch:
        for ship in SHIPS:
            batch.add_data_object(ship["props"], "Ship", ship["id"])

        batch.flush()

    yield client
    client.schema.delete_all()


def parse_client_options(request: FixtureRequest) -> dict:
    try:
        if isinstance(request.param, dict):
            return request.param
    except AttributeError:
        return


def test_get_data(client: weaviate.Client):
    """Test GraphQL's Get clause."""
    where_filter = {"path": ["size"], "operator": "LessThan", "valueInt": 10}
    result = client.query.get("Ship", ["name", "size"]).with_limit(2).with_where(where_filter).do()
    objects = get_objects_from_result(result)
    a_found = False
    d_found = False
    for obj in objects:
        if obj["name"] == "HMS British Name":
            a_found = True
        if obj["name"] == "Titanic":
            d_found = True
    assert a_found and d_found and len(objects) == 2


def test_get_data_with_where_contains_any(client: weaviate.Client):
    """Test GraphQL's Get clause with where filter."""
    where_filter = {"path": ["size"], "operator": "ContainsAny", "valueInt": [5]}
    result = client.query.get("Ship", ["name", "size"]).with_where(where_filter).do()
    objects = get_objects_from_result(result)
    assert len(objects) == 1 and objects[0]["name"] == "HMS British Name"


@pytest.mark.parametrize(
    "path,operator,value_type_key,value_type_value,name,expected_objects_len",
    [
        (
            ["description"],
            "ContainsAll",
            "valueString",
            ["sponges, sponges, sponges"],
            "The Crusty Crab",
            1,
        ),
        (
            ["description"],
            "ContainsAll",
            "valueText",
            ["sponges", "sponges", "sponges"],
            "The Crusty Crab",
            1,
        ),
        (
            ["description"],
            "ContainsAll",
            "valueStringArray",
            ["sponges", "sponges", "sponges"],
            "The Crusty Crab",
            1,
        ),
        (
            ["description"],
            "ContainsAll",
            "valueTextArray",
            ["sponges, sponges, sponges"],
            "The Crusty Crab",
            1,
        ),
        (
            ["description"],
            "ContainsAll",
            "valueStringList",
            ["sponges", "sponges", "sponges"],
            "The Crusty Crab",
            1,
        ),
        (
            ["description"],
            "ContainsAll",
            "valueTextList",
            ["sponges", "sponges", "sponges"],
            "The Crusty Crab",
            1,
        ),
        (
            ["description"],
            "ContainsAny",
            "valueString",
            ["sponges, sponges, sponges"],
            "The Crusty Crab",
            1,
        ),
        (
            ["description"],
            "ContainsAny",
            "valueText",
            ["sponges", "sponges", "sponges"],
            "The Crusty Crab",
            1,
        ),
        (
            ["description"],
            "ContainsAny",
            "valueStringArray",
            ["sponges", "sponges", "sponges"],
            "The Crusty Crab",
            1,
        ),
        (
            ["description"],
            "ContainsAny",
            "valueTextArray",
            ["sponges, sponges, sponges"],
            "The Crusty Crab",
            1,
        ),
        (
            ["description"],
            "ContainsAny",
            "valueStringList",
            ["sponges", "sponges", "sponges"],
            "The Crusty Crab",
            1,
        ),
        (
            ["description"],
            "ContainsAny",
            "valueTextList",
            ["sponges", "sponges", "sponges"],
            "The Crusty Crab",
            1,
        ),
        (["size"], "ContainsAll", "valueInt", [5], "HMS British Name", 1),
        (["size"], "ContainsAll", "valueIntList", [5], "HMS British Name", 1),
        (["size"], "ContainsAll", "valueIntArray", [5], "HMS British Name", 1),
        (["size"], "ContainsAny", "valueInt", [5], "HMS British Name", 1),
        (["size"], "ContainsAny", "valueIntList", [5], "HMS British Name", 1),
        (["size"], "ContainsAny", "valueIntArray", [5], "HMS British Name", 1),
        (["rating"], "ContainsAll", "valueNumber", [6.66], "The dragon ship", 1),
        (["rating"], "ContainsAll", "valueNumberList", [6.66], "The dragon ship", 1),
        (["rating"], "ContainsAll", "valueNumberArray", [6.66], "The dragon ship", 1),
        (["rating"], "ContainsAny", "valueNumber", [6.66], "The dragon ship", 1),
        (["rating"], "ContainsAny", "valueNumberList", [6.66], "The dragon ship", 1),
        (["rating"], "ContainsAny", "valueNumberArray", [6.66], "The dragon ship", 1),
        (["size"], "Equal", "valueInt", 5, "HMS British Name", 1),
        (["size"], "LessThan", "valueInt", 5, "Titanic", 1),
        (["size"], "LessThanEqual", "valueInt", 1, "Titanic", 1),
        (["size"], "GreaterThan", "valueInt", 300, "The Crusty Crab", 1),
        (["size"], "GreaterThanEqual", "valueInt", 303, "The Crusty Crab", 1),
        (["description"], "Like", "valueString", "sponges", "The Crusty Crab", 1),
        (["description"], "Like", "valueText", "sponges", "The Crusty Crab", 1),
        (["rating"], "IsNull", "valueBoolean", True, "irrelevant", 0),
        (["rating"], "NotEqual", "valueNumber", 6.66, "irrelevant", 5),
    ],
)
def test_get_data_with_where(
    client: weaviate.Client,
    path: List[str],
    operator: str,
    value_type_key: str,
    value_type_value: Union[List[int], List[str]],
    name,
    expected_objects_len: int,
):
    """Test GraphQL's Get clause with where filter."""
    where_filter = {
        "path": path,
        "operator": operator,
        value_type_key: value_type_value,
    }
    result = client.query.get("Ship", ["name"]).with_where(where_filter).do()
    objects = get_objects_from_result(result)
    if expected_objects_len == 0:
        assert objects is None
    else:
        assert len(objects) == expected_objects_len
        if expected_objects_len == 1:
            assert objects[0]["name"] == name


def test_get_data_after(client: weaviate.Client):
    full_results = client.query.get("Ship", ["name"]).with_additional(["id"]).do()
    for i, ship in enumerate(full_results["data"]["Get"]["Ship"][:-1]):
        result = (
            client.query.get("Ship", ["name"])
            .with_additional(["id"])
            .with_limit(1)
            .with_after(ship["_additional"]["id"])
            .do()
        )
        assert (
            result["data"]["Get"]["Ship"][0]["_additional"]["id"]
            == full_results["data"]["Get"]["Ship"][i + 1]["_additional"]["id"]
        )


def test_get_data_after_wrong_types(client: weaviate.Client):
    with pytest.raises(TypeError):
        client.query.get("Ship", ["name"]).with_additional(["id"]).with_limit(1).with_after(
            1234
        ).do()


def test_multi_get_data(client: weaviate.Client, people_schema):
    """Test GraphQL's MultiGet clause."""
    client.schema.create(people_schema)
    client.data_object.create(
        {
            "name": "John",
        },
        "Person",
    )
    result = client.query.multi_get(
        [
            client.query.get("Ship", ["name"])
            .with_alias("one")
            .with_sort({"path": ["name"], "order": "asc"}),
            client.query.get("Ship", ["size"])
            .with_alias("two")
            .with_sort({"path": ["size"], "order": "asc"}),
            client.query.get("Person", ["name"]),
        ]
    ).do()["data"]["Get"]
    assert result["one"][0]["name"] == "Artemis"
    assert result["two"][0]["size"] == 1
    assert result["Person"][0]["name"] == "John"


def test_aggregate_data(client: weaviate.Client):
    """Test GraphQL's Aggregate clause."""
    where_filter = {"path": ["name"], "operator": "Equal", "valueString": "The dragon ship"}

    result = (
        client.query.aggregate("Ship")
        .with_where(where_filter)
        .with_group_by_filter(["name"])
        .with_fields("groupedBy {value}")
        .with_fields("name{count}")
        .do()
    )

    aggregation = get_aggregation_from_aggregate_result(result)
    assert "groupedBy" in aggregation, "Missing groupedBy"
    assert "name" in aggregation, "Missing name property"


def test_aggregate_data_with_group_by_and_limit(client: weaviate.Client):
    """Test GraphQL's Aggregate clause with group_by and limit."""
    result = (
        client.query.aggregate("Ship")
        .with_fields("name{count}")
        .with_limit(2)
        .with_group_by_filter(["name"])
        .do()
    )

    objects = get_objects_from_aggregate_result(result)
    assert len(objects) == 2, "Expected 2 results"


def test_aggregate_data_with_just_limit(client: weaviate.Client):
    """Test GraphQL's Aggregate clause with only limit. It's idempotent."""
    result = client.query.aggregate("Ship").with_fields("name{count}").with_limit(2).do()

    objects = get_objects_from_aggregate_result(result)
    assert objects == [
        {"name": {"count": len(SHIPS)}}
    ], f"Expected only meta count for {len(SHIPS)} results"


def get_objects_from_result(result):
    return result["data"]["Get"]["Ship"]


def get_aggregation_from_aggregate_result(result):
    return result["data"]["Aggregate"]["Ship"][0]


def get_objects_from_aggregate_result(result):
    return result["data"]["Aggregate"]["Ship"]


@pytest.mark.parametrize("query", ["sponges", "sponges\n"])
def test_bm25(client: weaviate.Client, query):
    result = client.query.get("Ship", ["name"]).with_bm25(query, ["name", "description"]).do()
    assert len(result["data"]["Get"]["Ship"]) == 1
    assert result["data"]["Get"]["Ship"][0]["name"] == "The Crusty Crab"


def test_bm25_no_result(client: weaviate.Client):
    result = client.query.get("Ship", ["name"]).with_bm25("sponges\n", ["name"]).do()
    assert len(result["data"]["Get"]["Ship"]) == 0


@pytest.mark.parametrize("query", ["sponges", "sponges\n"])
@pytest.mark.parametrize("fusion_type", [HybridFusion.RANKED, HybridFusion.RELATIVE_SCORE, None])
def test_hybrid(client: weaviate.Client, query: str, fusion_type: Optional[HybridFusion]):
    """Test hybrid search with alpha=0.5 to have a combination of BM25 and vector search."""
    result = (
        client.query.get("Ship", ["name", "description"])
        .with_hybrid(query, alpha=0.5, fusion_type=fusion_type)
        .do()
    )

    # will find more results. "The Crusty Crab" is still first, because it matches with the BM25 search
    assert len(result["data"]["Get"]["Ship"]) >= 1
    assert result["data"]["Get"]["Ship"][0]["name"] == "The Crusty Crab"


@pytest.mark.parametrize(
    "properties,num_results",
    [(None, 1), ([], 1), (["description"], 1), (["description", "name"], 1), (["name"], 0)],
)
def test_hybrid_properties(
    client: weaviate.Client, properties: Optional[List[str]], num_results: int
):
    """Test hybrid search with alpha=0.5 to have a combination of BM25 and vector search."""
    result = (
        client.query.get("Ship", ["name"])
        .with_hybrid("sponges", alpha=0.0, properties=properties)
        .do()
    )

    # "The Crusty Crab" has "sponges" in its description, it cannot be found in other properties
    if num_results > 0:
        assert len(result["data"]["Get"]["Ship"]) >= 1

        assert result["data"]["Get"]["Ship"][0]["name"] == "The Crusty Crab"
    else:
        assert len(result["data"]["Get"]["Ship"]) == 0


@pytest.mark.parametrize("autocut,num_results", [(1, 1), (2, 6), (-1, len(SHIPS))])
def test_autocut(client: weaviate.Client, autocut, num_results):
    result = (
        client.query.get("Ship", ["name"])
        .with_hybrid(query="sponges", properties=["name", "description"], alpha=0.5)
        .with_autocut(autocut)
        .with_limit(len(SHIPS))
        .do()
    )
    assert len(result["data"]["Get"]["Ship"]) == num_results
    assert result["data"]["Get"]["Ship"][0]["name"] == "The Crusty Crab"


def test_group_by(client: weaviate.Client, people_schema):
    """Test hybrid search with alpha=0.5 to have a combination of BM25 and vector search."""
    client.schema.delete_all()
    client.schema.create(people_schema)

    persons = []
    for i in range(10):
        persons.append(uuid.uuid4())
        client.data_object.create({"name": "randomName" + str(i)}, "Person", persons[-1])

    client.data_object.create({}, "Call", "3ab05e06-2bb2-41d1-b5c5-e044f3aa9623")
    client.data_object.create({}, "Call", "3ab05e06-2bb2-41d1-b5c5-e044f3aa9622")

    # create refs
    for i in range(5):
        client.data_object.reference.add(
            to_uuid=persons[i],
            from_property_name="caller",
            from_uuid="3ab05e06-2bb2-41d1-b5c5-e044f3aa9623"
            if i % 2 == 0
            else "3ab05e06-2bb2-41d1-b5c5-e044f3aa9622",
            from_class_name="Call",
            to_class_name="Person",
        )

    result = (
        client.query.get("Call", ["caller{... on Person{name}}"])
        .with_near_object({"id": "3ab05e06-2bb2-41d1-b5c5-e044f3aa9622"})
        .with_group_by(properties=["caller"], groups=2, objects_per_group=3)
        .with_additional("group{hits {_additional{vector}caller{... on Person{name}}}}")
        .do()
    )

    # will find more results. "The Crusty Crab" is still first, because it matches with the BM25 search
    assert len(result["data"]["Get"]["Call"]) >= 1
    assert result["data"]["Get"]["Call"][0]["caller"][0]["name"] == "randomName0"


@pytest.mark.parametrize(
    "client,level",
    [
        ({"cluster": True}, ConsistencyLevel.ONE),
        ({"cluster": True}, ConsistencyLevel.QUORUM),
        ({"cluster": True}, ConsistencyLevel.ALL),
    ],
    indirect=["client"],
)
def test_consistency_level(client: weaviate.Client, level):
    result = (
        client.query.get("Ship", ["name"])
        .with_consistency_level(level)
        .with_additional("isConsistent")
        .do()
    )
    for _, res in enumerate(get_objects_from_result(result)):
        assert res["_additional"]["isConsistent"]


@pytest.mark.parametrize(
    "single,grouped",
    [
        ("Describe the following as a Facebook Ad: {review}", None),
        (None, "Describe the following as a LinkedIn Ad: {review}"),
        (
            "Describe the following as a Twitter Ad: {review}",
            "Describe the following as a Mastodon Ad: {review}",
        ),
        (
            "Describe the following as a Twitter Ad: \n Review: {review} \n Name: {name}",
            "Describe the following as a Mastodon Ad:  \n Review: {review} \n Name: {name}",
        ),
    ],
)
def test_generative_openai(single: str, grouped: str):
    """Test client credential flow with various providers."""
    api_key = os.environ.get("OPENAI_APIKEY")
    if api_key is None:
        pytest.skip("No OpenAI API key found.")

    client = weaviate.Client(
        "http://localhost:8086", additional_headers={"X-OpenAI-Api-Key": api_key}
    )
    client.schema.delete_all()
    wine_class = {
        "class": "Wine",
        "properties": [
            {"name": "name", "dataType": ["string"]},
            {"name": "review", "dataType": ["string"]},
        ],
        "moduleConfig": {"generative-openai": {}},
    }
    client.schema.create_class(wine_class)
    client.data_object.create(
        data_object={"name": "Super expensive wine", "review": "Tastes like a fresh ocean breeze"},
        class_name="Wine",
    )
    client.data_object.create(
        data_object={"name": "cheap wine", "review": "Tastes like forest"}, class_name="Wine"
    )

    result = (
        client.query.get("Wine", ["name", "review"])
        .with_generate(single_prompt=single, grouped_task=grouped)
        .do()
    )
    assert result["data"]["Get"]["Wine"][0]["_additional"]["generate"]["error"] is None

    grouped_properties = ["review"]
    result = (
        client.query.get("Wine", ["name", "review"])
        .with_generate(
            single_prompt=single, grouped_task=grouped, grouped_properties=grouped_properties
        )
        .do()
    )
    assert result["data"]["Get"]["Wine"][0]["_additional"]["generate"]["error"] is None


def test_graphql_with_tenant():
    client = weaviate.Client("http://localhost:8080")
    client.schema.delete_all()
    schema_class = {
        "class": "GraphQlTenantClass",
        "vectorizer": "none",
        "multiTenancyConfig": {"enabled": True},
    }

    tenants = ["tenant1", "tenant2"]
    client.schema.create_class(schema_class)
    client.schema.add_class_tenants(schema_class["class"], [Tenant(tenant) for tenant in tenants])

    nr_objects = 101
    with client.batch() as batch:
        for i in range(nr_objects):
            batch.add_data_object(
                class_name=schema_class["class"], tenant=tenants[i % 2], data_object={}
            )

    # no results without tenant
    results = client.query.get(schema_class["class"]).with_additional("id").do()
    assert results["data"]["Get"][schema_class["class"]] is None
    assert results["errors"] is not None

    # get call with tenant only returns the objects for a given tenant
    results = (
        client.query.get(schema_class["class"]).with_additional("id").with_tenant(tenants[0]).do()
    )
    assert len(results["data"]["Get"][schema_class["class"]]) == nr_objects // 2 + 1

    results = (
        client.query.get(schema_class["class"]).with_additional("id").with_tenant(tenants[1]).do()
    )
    assert len(results["data"]["Get"][schema_class["class"]]) == nr_objects // 2

    results = client.query.aggregate(schema_class["class"]).with_meta_count().do()
    assert results["data"]["Aggregate"][schema_class["class"]] is None
    assert results["errors"] is not None

    results = (
        client.query.aggregate(schema_class["class"]).with_meta_count().with_tenant(tenants[0]).do()
    )
    assert (
        int(results["data"]["Aggregate"][schema_class["class"]][0]["meta"]["count"])
        == nr_objects // 2 + 1
    )

    results = (
        client.query.aggregate(schema_class["class"]).with_meta_count().with_tenant(tenants[1]).do()
    )
    assert (
        int(results["data"]["Aggregate"][schema_class["class"]][0]["meta"]["count"])
        == nr_objects // 2
    )


def test_graphql_with_nested_object():
    client = weaviate.Client("http://localhost:8080")
    client.schema.delete_all()
    client.schema.create_class(
        {
            "class": "NestedObjectClass",
            "vectorizer": "none",
            "properties": [
                {
                    "name": "nested",
                    "dataType": ["object"],
                    "nestedProperties": [
                        {
                            "name": "name",
                            "dataType": ["text"],
                        },
                        {
                            "name": "names",
                            "dataType": ["text[]"],
                        },
                        {
                            "name": "age",
                            "dataType": ["int"],
                        },
                        {
                            "name": "ages",
                            "dataType": ["int[]"],
                        },
                        {
                            "name": "weight",
                            "dataType": ["number"],
                        },
                        {
                            "name": "weights",
                            "dataType": ["number[]"],
                        },
                        {
                            "name": "isAlive",
                            "dataType": ["boolean"],
                        },
                        {
                            "name": "areAlive",
                            "dataType": ["boolean[]"],
                        },
                        {
                            "name": "date",
                            "dataType": ["date"],
                        },
                        {
                            "name": "dates",
                            "dataType": ["date[]"],
                        },
                        {
                            "name": "uuid",
                            "dataType": ["uuid"],
                        },
                        {
                            "name": "uuids",
                            "dataType": ["uuid[]"],
                        },
                    ],
                }
            ],
        }
    )
    data = {
        "name": "nested object",
        "names": ["nested", "object"],
        "age": 42,
        "ages": [42, 43],
        "weight": 42.42,
        "weights": [42.42, 43.43],
        "isAlive": True,
        "areAlive": [True, False],
        "date": "2021-01-01T00:00:00Z",
        "dates": ["2021-01-01T00:00:00Z", "2021-01-02T00:00:00Z"],
        "uuid": "00000000-0000-0000-0000-000000000000",
        "uuids": ["00000000-0000-0000-0000-000000000000", "00000000-0000-0000-0000-000000000001"],
    }
    uuid_ = client.data_object.create({"nested": data}, "NestedObjectClass")

    results = client.query.get(
        "NestedObjectClass",
        [
            "nested { name names age ages weight weights isAlive areAlive date dates uuid uuids } _additional { id }"
        ],
    ).do()
    print(results)
    assert results["data"]["Get"]["NestedObjectClass"][0]["nested"] == data
    assert results["data"]["Get"]["NestedObjectClass"][0]["_additional"]["id"] == uuid_



from typing import List, Optional

from weaviate.collections.aggregations.base import _Aggregate
from weaviate.collections.classes.aggregate import (
    PropertiesMetrics,
    _AggregateReturn,
    _AggregateGroupByReturn,
)
from weaviate.collections.classes.filters import _Filters


class _NearVector(_Aggregate):
    def near_vector(
        self,
        near_vector: List[float],
        certainty: Optional[float] = None,
        distance: Optional[float] = None,
        object_limit: Optional[int] = None,
        filters: Optional[_Filters] = None,
        limit: Optional[int] = None,
        total_count: bool = False,
        return_metrics: Optional[PropertiesMetrics] = None,
    ) -> _AggregateReturn:
        """Aggregate metrics over the objects returned by a near vector search on this collection.

        At least one of `certainty`, `distance`, or `object_limit` must be specified here for the vector search.

        This method requires that the objects in the collection have associated vectors.

        Arguments:
            `near_vector`
                The vector to search on.
            `certainty`
                The minimum certainty of the vector search.
            `distance`
                The maximum distance of the vector search.
            `object_limit`
                The maximum number of objects to return from the vector search prior to the aggregation.
            `filters`
                The filters to apply to the search.
            `limit`
                The maximum number of aggregated objects to return.
            `total_count`
                Whether to include the total number of objects that match the query in the response.
            `return_metrics`
                A list of property metrics to aggregate together after the text search.

        Returns:
            A `_AggregateReturn` object that includes the aggregation objects.

        Raises:
            `weaviate.exceptions.WeaviateQueryException`:
                If an error occurs while performing the query against Weaviate.
        """
        builder = self._base(return_metrics, filters, limit, total_count)
        builder = self._add_near_vector(builder, near_vector, certainty, distance, object_limit)
        res = self._do(builder)
        return self._to_aggregate_result(res, return_metrics)


class _NearVectorGroupBy(_Aggregate):
    def near_vector(
        self,
        near_vector: List[float],
        group_by: str,
        certainty: Optional[float] = None,
        distance: Optional[float] = None,
        object_limit: Optional[int] = None,
        filters: Optional[_Filters] = None,
        limit: Optional[int] = None,
        total_count: bool = False,
        return_metrics: Optional[PropertiesMetrics] = None,
    ) -> List[_AggregateGroupByReturn]:
        """Aggregate metrics over the objects returned by a near vector search on this collection and grouping the results grouping the results by a property.

        At least one of `certainty`, `distance`, or `object_limit` must be specified here for the vector search.

        This method requires that the objects in the collection have associated vectors.

        Arguments:
            `near_vector`
                The vector to search on.
            `group_by`
                The property name to group the aggregation by.
            `certainty`
                The minimum certainty of the vector search.
            `distance`
                The maximum distance of the vector search.
            `object_limit`
                The maximum number of objects to return from the vector search prior to the aggregation.
            `filters`
                The filters to apply to the search.
            `limit`
                The maximum number of aggregated objects to return.
            `total_count`
                Whether to include the total number of objects that match the query in the response.
            `return_metrics`
                A list of property metrics to aggregate together after the text search.

        Returns:
            A `_AggregateGroupByReturn` object that includes the aggregation objects.

        Raises:
            `weaviate.exceptions.WeaviateQueryException`:
                If an error occurs while performing the query against Weaviate.
        """
        builder = (
            self._base(return_metrics, filters, limit, total_count)
            .with_group_by_filter([group_by])
            .with_fields(" groupedBy { path value } ")
        )
        builder = self._add_near_vector(builder, near_vector, certainty, distance, object_limit)
        res = self._do(builder)
        return self._to_group_by_result(res, return_metrics)



from google.protobuf import struct_pb2 as _struct_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class ConsistencyLevel(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    CONSISTENCY_LEVEL_UNSPECIFIED: _ClassVar[ConsistencyLevel]
    CONSISTENCY_LEVEL_ONE: _ClassVar[ConsistencyLevel]
    CONSISTENCY_LEVEL_QUORUM: _ClassVar[ConsistencyLevel]
    CONSISTENCY_LEVEL_ALL: _ClassVar[ConsistencyLevel]

CONSISTENCY_LEVEL_UNSPECIFIED: ConsistencyLevel
CONSISTENCY_LEVEL_ONE: ConsistencyLevel
CONSISTENCY_LEVEL_QUORUM: ConsistencyLevel
CONSISTENCY_LEVEL_ALL: ConsistencyLevel

class NumberArrayProperties(_message.Message):
    __slots__ = ["values", "prop_name"]
    VALUES_FIELD_NUMBER: _ClassVar[int]
    PROP_NAME_FIELD_NUMBER: _ClassVar[int]
    values: _containers.RepeatedScalarFieldContainer[float]
    prop_name: str
    def __init__(
        self, values: _Optional[_Iterable[float]] = ..., prop_name: _Optional[str] = ...
    ) -> None: ...

class IntArrayProperties(_message.Message):
    __slots__ = ["values", "prop_name"]
    VALUES_FIELD_NUMBER: _ClassVar[int]
    PROP_NAME_FIELD_NUMBER: _ClassVar[int]
    values: _containers.RepeatedScalarFieldContainer[int]
    prop_name: str
    def __init__(
        self, values: _Optional[_Iterable[int]] = ..., prop_name: _Optional[str] = ...
    ) -> None: ...

class TextArrayProperties(_message.Message):
    __slots__ = ["values", "prop_name"]
    VALUES_FIELD_NUMBER: _ClassVar[int]
    PROP_NAME_FIELD_NUMBER: _ClassVar[int]
    values: _containers.RepeatedScalarFieldContainer[str]
    prop_name: str
    def __init__(
        self, values: _Optional[_Iterable[str]] = ..., prop_name: _Optional[str] = ...
    ) -> None: ...

class BooleanArrayProperties(_message.Message):
    __slots__ = ["values", "prop_name"]
    VALUES_FIELD_NUMBER: _ClassVar[int]
    PROP_NAME_FIELD_NUMBER: _ClassVar[int]
    values: _containers.RepeatedScalarFieldContainer[bool]
    prop_name: str
    def __init__(
        self, values: _Optional[_Iterable[bool]] = ..., prop_name: _Optional[str] = ...
    ) -> None: ...

class ObjectPropertiesValue(_message.Message):
    __slots__ = [
        "non_ref_properties",
        "number_array_properties",
        "int_array_properties",
        "text_array_properties",
        "boolean_array_properties",
        "object_properties",
        "object_array_properties",
    ]
    NON_REF_PROPERTIES_FIELD_NUMBER: _ClassVar[int]
    NUMBER_ARRAY_PROPERTIES_FIELD_NUMBER: _ClassVar[int]
    INT_ARRAY_PROPERTIES_FIELD_NUMBER: _ClassVar[int]
    TEXT_ARRAY_PROPERTIES_FIELD_NUMBER: _ClassVar[int]
    BOOLEAN_ARRAY_PROPERTIES_FIELD_NUMBER: _ClassVar[int]
    OBJECT_PROPERTIES_FIELD_NUMBER: _ClassVar[int]
    OBJECT_ARRAY_PROPERTIES_FIELD_NUMBER: _ClassVar[int]
    non_ref_properties: _struct_pb2.Struct
    number_array_properties: _containers.RepeatedCompositeFieldContainer[NumberArrayProperties]
    int_array_properties: _containers.RepeatedCompositeFieldContainer[IntArrayProperties]
    text_array_properties: _containers.RepeatedCompositeFieldContainer[TextArrayProperties]
    boolean_array_properties: _containers.RepeatedCompositeFieldContainer[BooleanArrayProperties]
    object_properties: _containers.RepeatedCompositeFieldContainer[ObjectProperties]
    object_array_properties: _containers.RepeatedCompositeFieldContainer[ObjectArrayProperties]
    def __init__(
        self,
        non_ref_properties: _Optional[_Union[_struct_pb2.Struct, _Mapping]] = ...,
        number_array_properties: _Optional[
            _Iterable[_Union[NumberArrayProperties, _Mapping]]
        ] = ...,
        int_array_properties: _Optional[_Iterable[_Union[IntArrayProperties, _Mapping]]] = ...,
        text_array_properties: _Optional[_Iterable[_Union[TextArrayProperties, _Mapping]]] = ...,
        boolean_array_properties: _Optional[
            _Iterable[_Union[BooleanArrayProperties, _Mapping]]
        ] = ...,
        object_properties: _Optional[_Iterable[_Union[ObjectProperties, _Mapping]]] = ...,
        object_array_properties: _Optional[
            _Iterable[_Union[ObjectArrayProperties, _Mapping]]
        ] = ...,
    ) -> None: ...

class ObjectArrayProperties(_message.Message):
    __slots__ = ["values", "prop_name"]
    VALUES_FIELD_NUMBER: _ClassVar[int]
    PROP_NAME_FIELD_NUMBER: _ClassVar[int]
    values: _containers.RepeatedCompositeFieldContainer[ObjectPropertiesValue]
    prop_name: str
    def __init__(
        self,
        values: _Optional[_Iterable[_Union[ObjectPropertiesValue, _Mapping]]] = ...,
        prop_name: _Optional[str] = ...,
    ) -> None: ...

class ObjectProperties(_message.Message):
    __slots__ = ["value", "prop_name"]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    PROP_NAME_FIELD_NUMBER: _ClassVar[int]
    value: ObjectPropertiesValue
    prop_name: str
    def __init__(
        self,
        value: _Optional[_Union[ObjectPropertiesValue, _Mapping]] = ...,
        prop_name: _Optional[str] = ...,
    ) -> None: ...



import datetime
import io
import pathlib
import re
import sys
from typing import (
    Any,
    Dict,
    Generic,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)
from typing_extensions import is_typeddict

if sys.version_info < (3, 9):
    from typing_extensions import Annotated, get_type_hints, get_origin
else:
    from typing import Annotated, get_type_hints, get_origin

import uuid as uuid_lib

from google.protobuf import struct_pb2

from weaviate.collections.classes.config import ConsistencyLevel
from weaviate.collections.classes.grpc import (
    FromReference,
    MetadataQuery,
    FromNested,
    PROPERTIES,
)
from weaviate.collections.classes.internal import (
    _GroupByObject,
    _MetadataResult,
    _GenerativeObject,
    _Object,
    _Reference,
    _extract_property_type_from_annotated_reference,
    _extract_property_type_from_reference,
    _extract_properties_from_data_model,
    _GenerativeReturn,
    _GroupByResult,
    _GroupByReturn,
    _QueryReturn,
    GenerativeReturn,
    GroupByReturn,
    QueryReturn,
    ReturnProperties,
)
from weaviate.collections.classes.types import (
    Properties,
    TProperties,
)
from weaviate.collections.grpc.query import _QueryGRPC, GroupByResult, SearchResponse, SearchResult
from weaviate.connect import Connection
from weaviate.exceptions import WeaviateGrpcUnavailable
from weaviate.util import file_encoder_b64
from weaviate.proto.v1 import base_pb2, search_get_pb2

T = TypeVar("T")


class _Grpc(Generic[Properties]):
    def __init__(
        self,
        connection: Connection,
        name: str,
        consistency_level: Optional[ConsistencyLevel],
        tenant: Optional[str],
        type_: Optional[Type[Properties]],
    ):
        self.__connection = connection
        self.__name = name
        self.__tenant = tenant
        self.__consistency_level = consistency_level
        self._type = type_

    def _query(self) -> _QueryGRPC:
        if not self.__connection._grpc_available:
            raise WeaviateGrpcUnavailable()
        return _QueryGRPC(self.__connection, self.__name, self.__tenant, self.__consistency_level)

    @staticmethod
    def _extract_metadata_for_object(
        add_props: "search_get_pb2.MetadataResult",
    ) -> _MetadataResult:
        return _MetadataResult(
            uuid=uuid_lib.UUID(add_props.id) if len(add_props.id) > 0 else None,
            vector=[float(num) for num in add_props.vector] if len(add_props.vector) > 0 else None,
            distance=add_props.distance if add_props.distance_present else None,
            certainty=add_props.certainty if add_props.certainty_present else None,
            creation_time_unix=add_props.creation_time_unix
            if add_props.creation_time_unix_present
            else None,
            last_update_time_unix=add_props.last_update_time_unix
            if add_props.last_update_time_unix_present
            else None,
            score=add_props.score if add_props.score_present else None,
            explain_score=add_props.explain_score if add_props.explain_score_present else None,
            is_consistent=add_props.is_consistent,
            generative=add_props.generative if add_props.generative_present else None,
        )

    def _deserialize_primitive(self, value: Any, type_value: Any) -> Any:
        if type_value == uuid_lib.UUID:
            return uuid_lib.UUID(value)
        if type_value == datetime.datetime:
            return datetime.datetime.fromisoformat(value)
        if isinstance(type_value, list):
            return [
                self._deserialize_primitive(val, type_value[idx]) for idx, val in enumerate(value)
            ]
        if isinstance(value, struct_pb2.Struct):
            raise ValueError(
                f"The query returned an object value where it expected a primitive. Have you missed a NestedProperty specification in your query? {value}"
            )
        return value

    def __parse_nonref_properties_result(
        self,
        properties: Union[search_get_pb2.PropertiesResult, base_pb2.ObjectPropertiesValue],
        type_: Optional[Any],
    ) -> dict:
        hints = get_type_hints(type_) if get_origin(type_) is not dict and type_ is not None else {}
        result = {}

        for name, non_ref_prop in properties.non_ref_properties.items():
            result[name] = self._deserialize_primitive(non_ref_prop, hints.get(name))

        for number_array_property in properties.number_array_properties:
            result[number_array_property.prop_name] = [
                float(val) for val in number_array_property.values
            ]

        for int_array_property in properties.int_array_properties:
            result[int_array_property.prop_name] = [int(val) for val in int_array_property.values]

        for text_array_property in properties.text_array_properties:
            result[text_array_property.prop_name] = [str(val) for val in text_array_property.values]

        for boolean_array_property in properties.boolean_array_properties:
            result[boolean_array_property.prop_name] = [
                bool(val) for val in boolean_array_property.values
            ]

        for object_property in properties.object_properties:
            result[object_property.prop_name] = self.__parse_nonref_properties_result(
                object_property.value, type_=hints.get(object_property.prop_name)
            )

        for object_array_property in properties.object_array_properties:
            result[object_array_property.prop_name] = [
                self.__parse_nonref_properties_result(
                    object_property,
                    hints.get(
                        object_array_property.prop_name,
                        [None for _ in range(len(object_array_property.values))][i],
                    ),
                )
                for i, object_property in enumerate(object_array_property.values)
            ]

        return result

    def __parse_ref_properties_result(
        self, properties: "search_get_pb2.PropertiesResult", type_: Optional[Type[T]]
    ) -> dict:
        hints = get_type_hints(type_) if get_origin(type_) is not dict and type_ is not None else {}
        result = {}
        for ref_prop in properties.ref_props:
            hint = hints.get(ref_prop.prop_name)
            if hint is not None:
                if get_origin(hint) is Annotated:
                    referenced_property_type = _extract_property_type_from_annotated_reference(hint)
                else:
                    assert get_origin(hint) is _Reference
                    referenced_property_type = _extract_property_type_from_reference(hint)
                result[ref_prop.prop_name] = _Reference._from(
                    [
                        _Object(
                            properties=self.__parse_result(prop, referenced_property_type),
                            metadata=self._extract_metadata_for_object(prop.metadata)._to_return(),
                        )
                        for prop in ref_prop.properties
                    ]
                )
            else:
                result[ref_prop.prop_name] = _Reference._from(
                    [
                        _Object(
                            properties=self.__parse_result(prop, Dict[str, Any]),
                            metadata=self._extract_metadata_for_object(prop.metadata)._to_return(),
                        )
                        for prop in ref_prop.properties
                    ]
                )
        return result

    def __parse_result(
        self, properties: "search_get_pb2.PropertiesResult", type_: Optional[Type[T]]
    ) -> T:
        nonref_result = self.__parse_nonref_properties_result(properties, type_)
        ref_result = self.__parse_ref_properties_result(properties, type_)
        return cast(T, {**nonref_result, **ref_result})

    def __result_to_query_object(self, res: SearchResult, type_: Optional[Type[T]]) -> _Object[T]:
        properties = self.__parse_result(res.properties, type_)
        metadata = self._extract_metadata_for_object(res.metadata)
        return _Object[T](properties=properties, metadata=metadata._to_return())

    def __result_to_generative_object(
        self, res: SearchResult, type_: Optional[Type[T]]
    ) -> _GenerativeObject[T]:
        properties = self.__parse_result(res.properties, type_)
        metadata = self._extract_metadata_for_object(res.metadata)
        return _GenerativeObject[T](
            properties=properties, metadata=metadata._to_return(), generated=metadata.generative
        )

    def __result_to_group(self, res: GroupByResult, type_: Optional[Type[T]]) -> _GroupByResult[T]:
        return _GroupByResult[T](
            objects=[self.__result_to_query_object(obj, type_) for obj in res.objects],
            name=res.name,
            number_of_objects=res.number_of_objects,
            min_distance=res.min_distance,
            max_distance=res.max_distance,
        )

    def _result_to_query_return(
        self,
        res: SearchResponse,
        type_: Optional[ReturnProperties[TProperties]],
    ) -> QueryReturn[Properties, TProperties]:
        if is_typeddict(type_):
            type_ = cast(Type[TProperties], type_)  # we know it's a typeddict
            return _QueryReturn[TProperties](
                objects=[self.__result_to_query_object(obj, type_=type_) for obj in res.results]
            )
        else:
            return _QueryReturn[Properties](
                objects=[
                    self.__result_to_query_object(obj, type_=self._type) for obj in res.results
                ]
            )

    def _result_to_generative_return(
        self,
        res: SearchResponse,
        type_: Optional[ReturnProperties[TProperties]],
    ) -> GenerativeReturn[Properties, TProperties]:
        if is_typeddict(type_):
            type_ = cast(Type[TProperties], type_)  # we know it's a typeddict
            return _GenerativeReturn[TProperties](
                objects=[
                    self.__result_to_generative_object(obj, type_=type_) for obj in res.results
                ],
                generated=res.generative_grouped_result
                if res.generative_grouped_result != ""
                else None,
            )
        else:
            return _GenerativeReturn[Properties](
                objects=[
                    self.__result_to_generative_object(obj, type_=self._type) for obj in res.results
                ],
                generated=res.generative_grouped_result
                if res.generative_grouped_result != ""
                else None,
            )

    def _result_to_groupby_return(
        self,
        res: SearchResponse,
        type_: Optional[ReturnProperties[TProperties]],
    ) -> GroupByReturn[Properties, TProperties]:
        if is_typeddict(type_):
            type_ = cast(Type[TProperties], type_)  # we know it's a typeddict
            groups = {
                group.name: self.__result_to_group(group, type_) for group in res.group_by_results
            }
            objects_group_by = [
                _GroupByObject[TProperties](
                    properties=obj.properties, metadata=obj.metadata, belongs_to_group=group.name
                )
                for group in groups.values()
                for obj in group.objects
            ]
            return _GroupByReturn[TProperties](objects=objects_group_by, groups=groups)
        else:
            groupss = {
                group.name: self.__result_to_group(group, self._type)
                for group in res.group_by_results
            }
            objects_group_byy = [
                _GroupByObject[Properties](
                    properties=obj.properties, metadata=obj.metadata, belongs_to_group=group.name
                )
                for group in groupss.values()
                for obj in group.objects
            ]
            return _GroupByReturn[Properties](objects=objects_group_byy, groups=groupss)

    def __parse_generic_properties(
        self, generic_properties: Type[TProperties]
    ) -> Optional[PROPERTIES]:
        if not is_typeddict(generic_properties):
            raise TypeError(
                f"return_properties must only be a TypedDict or PROPERTIES within this context but is {type(generic_properties)}"
            )
        return _extract_properties_from_data_model(generic_properties)

    def __parse_properties(self, return_properties: Optional[PROPERTIES]) -> Optional[PROPERTIES]:
        return _PropertiesParser().parse(return_properties)

    def _parse_return_properties(
        self, return_properties: Optional[ReturnProperties[TProperties]]
    ) -> Tuple[Optional[PROPERTIES], Optional[MetadataQuery]]:
        if (
            isinstance(return_properties, list)
            or isinstance(return_properties, str)
            or isinstance(return_properties, FromReference)
            or isinstance(return_properties, FromNested)
            or (return_properties is None and self._type is None)
        ):
            return self.__parse_properties(return_properties), None
        elif return_properties is None and self._type is not None:
            return self.__parse_generic_properties(self._type), MetadataQuery._full()
        else:
            assert return_properties is not None
            return self.__parse_generic_properties(return_properties), None

    @staticmethod
    def _parse_media(media: Union[str, pathlib.Path, io.BufferedReader]) -> str:
        if isinstance(media, str):  # if already encoded by user
            return media
        else:
            return file_encoder_b64(media)


class _PropertiesParser:
    def __init__(self) -> None:
        self.__from_references_by_prop_name: Dict[str, FromReference] = {}
        self.__non_ref_properties: List[str] = []

    def parse(self, properties: Optional[PROPERTIES]) -> Optional[PROPERTIES]:
        if (
            properties is None
            or isinstance(properties, str)
            or isinstance(properties, FromReference)
            or isinstance(properties, FromNested)
        ):
            if isinstance(properties, str) and properties.startswith("__"):
                self.__parse_reference_property_string(properties)
                return list(self.__from_references_by_prop_name.values())
            else:
                return properties
        elif isinstance(properties, list):
            for prop in properties:
                if prop is None:
                    continue
                if isinstance(prop, str):
                    if prop.startswith("__"):
                        self.__parse_reference_property_string(prop)
                    else:
                        self.__non_ref_properties.append(prop)
                elif isinstance(prop, FromReference):
                    self.__from_references_by_prop_name[prop.link_on] = prop
            return [*self.__non_ref_properties, *self.__from_references_by_prop_name.values()]
        else:
            raise TypeError(
                f"return_properties must be a list of strings and/or FromReferences, a string, or a FromReference but is {type(properties)}"
            )

    def __parse_reference_property_string_without_options(self, ref_prop: str) -> None:
        match = re.search(r"__([^_]+)", ref_prop)
        if match is None:
            raise ValueError(
                f"return reference property {ref_prop} must be in the format __{{prop_name}} or __{{prop_name}}__{{properties|metadata}}_{{nested_prop_name}} when using string syntax"
            )
        else:
            prop_name = match.group(1)
            existing_from_reference = self.__from_references_by_prop_name.get(prop_name)
            if existing_from_reference is None:
                self.__from_references_by_prop_name[prop_name] = FromReference(link_on=prop_name)

    def __parse_reference_property_string(self, ref_prop: str) -> None:
        match_ = re.search(r"__([^_]+)__([^_]+)__([\w_]+)", ref_prop)
        if match_ is None:
            self.__parse_reference_property_string_without_options(ref_prop)
            return

        prop_name = match_.group(1)
        existing_from_reference = self.__from_references_by_prop_name.get(prop_name)
        properties_or_metadata = match_.group(2)
        if properties_or_metadata not in ["properties", "metadata"]:
            raise ValueError(
                f"return reference property {ref_prop} must be in the format __{{prop_name}} or __{{prop_name}}__{{properties|metadata}}_{{nested_prop_name}} when using string syntax"
            )
        nested_prop_name = match_.group(3)
        if existing_from_reference is None:
            self.__from_references_by_prop_name[prop_name] = FromReference(
                link_on=prop_name,
                return_properties=[nested_prop_name]
                if properties_or_metadata == "properties"
                else None,
                return_metadata=MetadataQuery(**{nested_prop_name: True})
                if properties_or_metadata == "metadata"
                else None,
            )
        else:
            if properties_or_metadata == "properties":
                if existing_from_reference.return_properties is None:
                    self.__from_references_by_prop_name[prop_name].return_properties = [
                        nested_prop_name
                    ]
                else:
                    assert isinstance(existing_from_reference.return_properties, list)
                    existing_from_reference.return_properties.append(nested_prop_name)
            else:
                if existing_from_reference.return_metadata is None:
                    metadata = MetadataQuery()
                else:
                    metadata = existing_from_reference.return_metadata
                setattr(metadata, nested_prop_name, True)
                self.__from_references_by_prop_name[prop_name].return_metadata = metadata
